from difflib import SequenceMatcher
from pathlib import Path, PurePosixPath
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING, Literal, Optional, Union, Sequence, Mapping, Callable, List

from dagshub_annotation_converter.converters.cvat import load_cvat_from_zip
from dagshub_annotation_converter.converters.yolo import load_yolo_from_fs
from dagshub_annotation_converter.formats.label_studio.task import LabelStudioTask
from dagshub_annotation_converter.formats.yolo import YoloContext
from dagshub_annotation_converter.ir.image.annotations.base import IRAnnotationBase

from dagshub.common.api import UserAPI
from dagshub.common.api.repo import PathNotFoundError
from dagshub.common.helpers import log_message

if TYPE_CHECKING:
    from dagshub.data_engine.model.datasource import Datasource

AnnotationType = Literal["yolo", "cvat"]
AnnotationLocation = Literal["repo", "disk"]


class AnnotationsNotFoundError(Exception):
    def __init__(self, path):
        super().__init__(f'Annotations not found at path "{path}" in neither disk or repository.')


class CannotRemapPathError(Exception):
    def __init__(self, a_path, b_path):
        super().__init__(f"Cannot map from path {a_path} to path {b_path}")


class AnnotationImporter:
    """
    Class for importing annotations into a datasource from different formats.
    """

    def __init__(
        self,
        ds: "Datasource",
        annotations_type: AnnotationType,
        annotations_file: Union[str, Path],
        load_from: Optional[AnnotationLocation] = None,
        **kwargs,
    ):
        self.ds = ds.__deepcopy__()
        self.ds.clear_query()
        self.annotations_type = annotations_type
        self.annotations_file = Path(annotations_file)
        self.load_from = load_from if load_from is not None else self.determine_load_location(ds, annotations_file)
        self.additional_args = kwargs

        if self.annotations_type == "yolo":
            if "yolo_type" not in kwargs:
                raise ValueError(
                    "YOLO type must be provided in the additional arguments. "
                    'Add `yolo_type="bbox"|"segmentation"|pose"` to the arguments.'
                )

    def import_annotations(self) -> Mapping[str, Sequence[IRAnnotationBase]]:
        # Double check that the annotation file exists
        if self.load_from == "disk":
            if not self.annotations_file.exists():
                raise AnnotationsNotFoundError(self.annotations_file)
        elif self.load_from == "repo":
            try:
                self.ds.source.repoApi.list_path(self.annotations_file.as_posix())
            except PathNotFoundError:
                raise AnnotationsNotFoundError(self.annotations_file)

        annotations_file = self.annotations_file

        with TemporaryDirectory() as tmp_dir:
            tmp_dir_path = Path(tmp_dir)
            if self.load_from == "repo":
                self.download_annotations(tmp_dir_path)
                annotations_file = tmp_dir_path / annotations_file.name

            # Convert annotations
            log_message("Loading annotations...")
            annotation_dict: Mapping[str, Sequence[IRAnnotationBase]]
            if self.annotations_type == "yolo":
                annotation_dict, _ = load_yolo_from_fs(
                    annotation_type=self.additional_args["yolo_type"], meta_file=annotations_file
                )
            elif self.annotations_type == "cvat":
                annotation_dict = load_cvat_from_zip(annotations_file)

            return annotation_dict

    def download_annotations(self, dest_dir: Path):
        log_message("Downloading annotations from repository")
        repoApi = self.ds.source.repoApi
        if self.annotations_type == "cvat":
            # Download just the annotation file
            repoApi.download(self.annotations_file.as_posix(), dest_dir, keep_source_prefix=True)
        elif self.annotations_type == "yolo":
            # Download the dataset .yaml file and the images + annotations
            # Download the file
            repoApi.download(self.annotations_file.as_posix(), dest_dir, keep_source_prefix=True)
            # Get the YOLO Context from the downloaded file
            meta_file = dest_dir / self.annotations_file
            context = YoloContext.from_yaml_file(meta_file, annotation_type=self.additional_args["yolo_type"])
            # Download the annotation data
            assert context.path is not None
            repoApi.download(self.annotations_file.parent / context.path, dest_dir, keep_source_prefix=True)

    @staticmethod
    def determine_load_location(ds: "Datasource", annotations_path: Union[str, Path]) -> AnnotationLocation:
        # Local files take priority
        if Path(annotations_path).exists():
            return "disk"

        # Try to find it in the repo otherwise
        try:
            files = ds.source.repoApi.list_path(Path(annotations_path).as_posix())
            if len(files) > 0:
                return "repo"
        except PathNotFoundError:
            pass

        # TODO: handle repo bucket too

        raise AnnotationsNotFoundError(annotations_path)

    def remap_annotations(
        self,
        annotations: Mapping[str, Sequence[IRAnnotationBase]],
        remap_func: Optional[Callable[[str], Optional[str]]] = None,
    ) -> Mapping[str, Sequence[IRAnnotationBase]]:
        """
        Remaps the filenames in the annotations to the datasource's data points.

        Args:
            annotations: Annotations to remap
            remap_func: Function that maps from an annotation path to a datapoint path. \
                If None, we try to guess it by getting a datapoint and remapping that path
        """
        if remap_func is None:
            first_ann = list(annotations.keys())[0]
            first_ann_filename = PurePosixPath(first_ann).name
            queried = self.ds["path"].endswith(first_ann_filename).select("size").all()
            dp_paths = [dp.path for dp in queried]
            remap_func = self.guess_annotation_filename_remapping(first_ann, dp_paths)

        remapped = {}

        for filename, anns in annotations.items():
            new_filename = remap_func(filename)
            if new_filename is None:
                log_message(
                    f'Skipping annotation with filename "{filename}" because it could not be mapped to a datapoint'
                )
                continue
            for ann in anns:
                assert ann.filename is not None
                ann.filename = remap_func(ann.filename)
            remapped[new_filename] = anns

        return remapped

    @staticmethod
    def guess_annotation_filename_remapping(
        annotation_path: str, datapoint_paths: List[str]
    ) -> Callable[[str], Optional[str]]:
        """
        Guesses the remapping function from the annotations to the data points.

        Args:
            annotation_path: path of an existing annotations
            datapoint_paths: paths of the data points in the datasource that end with the filename of this annotation
        """

        if len(datapoint_paths) == 0:
            raise ValueError(f"No datapoints found that match the annotation path {annotation_path}")

        dp_path = datapoint_paths[0]

        if len(datapoint_paths) > 1:
            # TODO: Maybe prompt user to choose a fitting datapoint (ordered by similarity)
            dp_path = AnnotationImporter.get_best_fit_datapoint_path(annotation_path, datapoint_paths)
            log_message(f'Multiple datapoints found for annotation path "{annotation_path}". Using "{dp_path}"')

        return AnnotationImporter.generate_path_map_func(annotation_path, dp_path)

    @staticmethod
    def generate_path_map_func(ann_path: str, dp_path: str) -> Callable[[str], Optional[str]]:
        ann_path_posix = PurePosixPath(ann_path)
        dp_path_posix = PurePosixPath(dp_path)

        matcher = SequenceMatcher(
            None,
            ann_path_posix.parts,
            dp_path_posix.parts,
        )
        diff = matcher.get_matching_blocks()

        # Make sure that both sequences have the same end, get the common part.
        # Then the rest is going to be the prefix that is either added or subtracted.

        # We need there to be only one match that is at the very end, otherwise we throw an error
        if len(diff) != 2:
            raise CannotRemapPathError(ann_path, dp_path)

        match = diff[0]
        # Make sure that the match goes until the end
        if match.a + match.size != len(ann_path_posix.parts) or match.b + match.size != len(dp_path_posix.parts):
            raise CannotRemapPathError(ann_path, dp_path)
        # ONE of the paths need to go until the start
        if match.a != 0 and match.b != 0:
            raise CannotRemapPathError(ann_path, dp_path)

        # If the match is total, just return identity
        if match.a == 0 and match.b == 0:

            def identity_func(x: str) -> str:
                return x.replace(ann_path, dp_path)

            return identity_func

        # The function that maps ends up being:
        # - Get the common part of the path
        # - Either remove the remainder, or add the prefix, depending on which is longer

        if match.b > match.a:
            # Add a prefix
            prefix = dp_path_posix.parts[match.a : match.b]

            def add_prefix(x: str) -> Optional[str]:
                return PurePosixPath(*prefix, x).as_posix()

            return add_prefix

        else:
            # Remove the prefix
            def remove_prefix(x: str) -> Optional[str]:
                p = PurePosixPath(x)
                if len(p.parts) <= match.a:
                    return None
                return PurePosixPath(*p.parts[match.a :]).as_posix()

            return remove_prefix

    @staticmethod
    def get_best_fit_datapoint_path(ann_path: str, datapoint_paths: List[str]) -> str:
        """
        Get the datapoint path that is the closest to the annotation path.

        Args:
            ann_path: path of an annotation
            datapoint_paths: paths of the data points in the datasource that end with the filename of this annotation
        """
        best_match: Optional[str] = None
        best_match_length: Optional[int] = None

        for dp_path in datapoint_paths:
            ann_path_posix = PurePosixPath(ann_path)
            dp_path_posix = PurePosixPath(dp_path)

            matcher = SequenceMatcher(
                None,
                ann_path_posix.parts,
                dp_path_posix.parts,
            )
            diff = matcher.get_matching_blocks()

            if len(diff) != 2:  # Has multiple matches - bad
                continue
            match = diff[0]
            if match.a != 0 and match.b != 0:
                continue

            # Exact match - perfect!
            if match.a == 0 and match.b == 0:
                return dp_path

            if best_match_length is None or match.size > best_match_length:
                best_match = dp_path
                best_match_length = match.size
        if best_match is None:
            raise ValueError(f"No good match found for annotation path {ann_path} in the datasource.")
        return best_match

    def convert_to_ls_tasks(self, annotations: Mapping[str, Sequence[IRAnnotationBase]]) -> Mapping[str, bytes]:
        """
        Converts the annotations to Label Studio tasks.
        """
        current_user_id = UserAPI.get_current_user(self.ds.source.repoApi.host).user_id
        tasks = {}
        for filename, anns in annotations.items():
            t = LabelStudioTask(user_id=current_user_id)
            t.data["image"] = self.ds.source.raw_path(filename)
            t.add_ir_annotations(anns)
            tasks[filename] = t.model_dump_json().encode("utf-8")
        return tasks



import datetime
from typing import Type, Dict, Set

import dacite

from dagshub.data_engine.client.models import IntegrationStatus, DatasourceType, PreprocessingStatus
from dagshub.data_engine.dtypes import MetadataFieldType

metadataTypeLookup = {
    int: MetadataFieldType.INTEGER,
    bool: MetadataFieldType.BOOLEAN,
    float: MetadataFieldType.FLOAT,
    str: MetadataFieldType.STRING,
    bytes: MetadataFieldType.BLOB,
    datetime.datetime: MetadataFieldType.DATETIME,
}

metadataTypeLookupReverse: Dict[str, Type] = {}
for k, v in metadataTypeLookup.items():
    metadataTypeLookupReverse[v.value] = k


def timestamp_to_datetime(timestamp: int) -> datetime.datetime:
    return datetime.datetime.fromtimestamp(timestamp, tz=datetime.timezone.utc)


dacite_config = dacite.Config(
    cast=[IntegrationStatus, DatasourceType, PreprocessingStatus, MetadataFieldType, Set],
    type_hooks={datetime.datetime: timestamp_to_datetime},
)



from dataclasses import dataclass
from typing import TYPE_CHECKING, List, Optional, Dict

from dagshub.data_engine.client.models import MetadataFieldSchema
from dagshub.data_engine.dtypes import MetadataFieldType

if TYPE_CHECKING:
    from dagshub.data_engine.model.datasource import DatapointMetadataUpdateEntry, Datasource

MAX_STRING_FIELD_LENGTH = 512


@dataclass
class UploadingMetadataInfo:
    field_name: str
    """Name of the field"""
    field_type: MetadataFieldType
    """Type of the metadata in the field"""
    existing_metadata_in_ds: Optional[MetadataFieldSchema] = None
    """Info about the existing field in the datasource. If None, this is a new field"""
    longest_value: str = ""
    """For string fields - longest value"""


class MultipleDataTypesUploadedError(Exception):
    def __init__(self, field_name: str, orig_type: MetadataFieldType, other_type: MetadataFieldType):
        self.field_name = field_name
        self.orig_type = orig_type
        self.other_type = other_type

    def __str__(self):
        return (
            f"Tried to upload data with different values for field {self.field_name}. "
            f"Original type is {self.orig_type.value}, tried to upload {self.other_type.value}"
        )


class StringFieldValueTooLongError(Exception):
    def __init__(self, field_name: str, value: str):
        self.field_name = field_name
        self.value = value

    def __str__(self):
        return (
            f"Trying to insert a value into a string field {self.field_name} "
            f"that is longer than the maximum allowed length of {MAX_STRING_FIELD_LENGTH}.\n"
            "If you need to upload large text, use a document field instead.\n"
            f"Value that was over the limit: {self.value}"
        )


def precalculate_metadata_info(
    ds: "Datasource", metadata: List["DatapointMetadataUpdateEntry"]
) -> Dict[str, UploadingMetadataInfo]:
    """
    Calculates information about uploading metadata + checks for accidental mixing of data types along the way
    """
    res = {}
    for m in metadata:
        if m.key not in res:
            info = UploadingMetadataInfo(
                field_name=m.key,
                field_type=m.valueType,
                existing_metadata_in_ds=next(filter(lambda f: f.name == m.key, ds.fields), None),
            )
            res[m.key] = info

            if info.existing_metadata_in_ds is not None:
                # Check that user is not uploading other type of value accidentally
                if info.existing_metadata_in_ds.valueType != m.valueType:
                    raise MultipleDataTypesUploadedError(m.key, info.existing_metadata_in_ds.valueType, m.valueType)

        else:
            info = res[m.key]

            if info.field_type != m.valueType:
                raise MultipleDataTypesUploadedError(m.key, info.field_type, m.valueType)

        if info.field_type == MetadataFieldType.STRING:
            info.longest_value = info.longest_value if len(info.longest_value) > len(m.value) else m.value
    return res


def validate_uploading_metadata(precalculated_info: Dict[str, UploadingMetadataInfo]):
    """
    Run client-side validations on uploading metadata

    Current validations:
        - String fields can't be too big, but only if it's being inserted into an existing field
            New fields are OK, they will be transformed into blobs later during the transforms
        - [Run during precalculation] Can't upload data of different types
        - [Run during precalculation] Can't upload data of a type different to the one already in datasource
    """
    for info in precalculated_info.values():
        if (
            info.field_type == MetadataFieldType.STRING
            and info.existing_metadata_in_ds is not None
            and len(info.longest_value) > MAX_STRING_FIELD_LENGTH
        ):
            raise StringFieldValueTooLongError(info.field_name, info.longest_value)



import logging
from typing import List, TYPE_CHECKING, Dict

from dagshub.common.helpers import log_message
from dagshub.data_engine import dtypes
from dagshub.data_engine.dtypes import MetadataFieldType
from dagshub.data_engine.model.metadata.util import wrap_bytes
from dagshub.data_engine.model.metadata.validation import UploadingMetadataInfo, MAX_STRING_FIELD_LENGTH

if TYPE_CHECKING:
    from dagshub.data_engine.model.datasource import Datasource, DatapointMetadataUpdateEntry

logger = logging.getLogger(__name__)


def _transform_strings_to_documents(ds: "Datasource", metadata: List["DatapointMetadataUpdateEntry"], field_name: str):
    log_message(f"Uploading field {field_name} as a document field due to long string values")
    ds.metadata_field(field_name).set_type(dtypes.Document).apply()
    for m in metadata:
        if m.key != field_name:
            continue
        m.valueType = MetadataFieldType.BLOB
        m.value = wrap_bytes(m.value.encode("utf-8"))


def run_preupload_transforms(
    ds: "Datasource",
    metadata: List["DatapointMetadataUpdateEntry"],
    precalculated_info: Dict[str, UploadingMetadataInfo],
):
    """
    Run transformations on metadata that should happen before uploading.

    Current transformations:
        - New string fields with large values get converted to a Document type
    """

    for info in precalculated_info.values():
        if (
            info.field_type == MetadataFieldType.STRING
            and info.existing_metadata_in_ds is None
            and len(info.longest_value) > MAX_STRING_FIELD_LENGTH
        ):
            _transform_strings_to_documents(ds, metadata, info.field_name)



import base64
import gzip


def wrap_bytes(val: bytes) -> str:
    """
    Handles bytes values for uploading metadata
    The process is gzip -> base64

    :meta private:
    """
    compressed = gzip.compress(val)
    return base64.b64encode(compressed).decode("utf-8")



import dataclasses
import logging
from typing import TYPE_CHECKING, Type, Union, Set, Optional, Literal

from dagshub.data_engine.client.models import MetadataFieldSchema
from dagshub.data_engine.dtypes import DagshubDataType, MetadataFieldType, ReservedTags, ThumbnailType
from dagshub.data_engine.model.schema_util import metadataTypeLookup

if TYPE_CHECKING:
    from dagshub.data_engine.model.datasource import Datasource

logger = logging.getLogger(__name__)


class MetadataFieldBuilder:
    """
    Builder class for changing properties of a metadata field in a datasource.
    It is also possible to create a new empty field with predefined schema with this builder.
    All functions return back the builder object to facilitate a builder pattern, for example::

        builder.set_type(bytes).set_annotation().apply()
    """

    def __init__(self, datasource: "Datasource", field_name: str):
        self.datasource = datasource

        self._field_name = field_name

        preexisting_schema = next(filter(lambda f: f.name == field_name, datasource.fields), None)

        # Make a copy of the dataclass, so we don't change the base schema
        if preexisting_schema is not None:
            preexisting_schema = dataclasses.replace(preexisting_schema)
            if preexisting_schema.tags is not None:
                preexisting_schema.tags = preexisting_schema.tags.copy()

        self._schema = preexisting_schema
        self.already_exists = self._schema is not None

    @property
    def schema(self) -> MetadataFieldSchema:
        if self._schema is None:
            raise RuntimeError(
                f"Field {self._field_name} is a new field. "
                "Make sure to set_type() the field before setting any other properties"
            )
        return self._schema

    def set_type(self, t: Union[Type, DagshubDataType]) -> "MetadataFieldBuilder":
        """
        Set the type of the field.
        The type can be either a Python primitive supported by the Data Engine
        (``str``, ``bool``, ``int``, ``float``, ``bytes``)
        or it can be a :class:`~dagshub.data_engine.dtypes.DagshubDataType` inheritor.
        The DataType inheritors can define additional tags on top of just the basic backing type
        """
        backing_type = self._get_backing_type(t)

        if self._schema is None:
            self._schema = MetadataFieldSchema(
                name=self._field_name, valueType=backing_type, multiple=False, tags=set()
            )
            if issubclass(t, DagshubDataType) and t.custom_tags is not None:
                self._schema.tags = t.custom_tags.copy()
        else:
            if backing_type != self._schema.valueType:
                raise ValueError(
                    "Can't change a type of an already existing field "
                    f"(changing from {self._schema.valueType.value} to {backing_type.value})"
                )
            if issubclass(t, DagshubDataType) and t.custom_tags is not None:
                self._add_tags(t.custom_tags)

        return self

    def set_annotation(self, is_annotation: bool = True) -> "MetadataFieldBuilder":
        """
        Mark or unmark the field as annotation field
        """
        self._set_or_unset(ReservedTags.ANNOTATION.value, is_annotation)
        return self

    def set_thumbnail(
        self,
        thumbnail_type: Optional[Literal["video", "audio", "image", "pdf", "text"]] = None,
        is_thumbnail: bool = True,
    ) -> "MetadataFieldBuilder":
        """
        Mark or unmark the field as thumbnail field, with the specified thumbnail type
        """

        # Remove thumbnail tag
        if not is_thumbnail:
            self._set_or_unset_thumbnails(ReservedTags.THUMBNAIL_VIZ.value, is_thumbnail)
            return self

        # Set thumbnail tag
        if thumbnail_type is None:
            raise ValueError("Thumbnail type must be specified")

        valid_types = ", ".join([t.value for t in ThumbnailType])
        try:
            thumbnail_type = ThumbnailType(thumbnail_type)
        except ValueError:
            raise ValueError(f"'{thumbnail_type}' is not a valid thumbnail type. Valid types are: {valid_types}")

        tag: ReservedTags

        if thumbnail_type == ThumbnailType.VIDEO:
            tag = ReservedTags.VIDEO
        elif thumbnail_type == ThumbnailType.AUDIO:
            tag = ReservedTags.AUDIO
        elif thumbnail_type == ThumbnailType.IMAGE:
            tag = ReservedTags.IMAGE
        elif thumbnail_type == ThumbnailType.PDF:
            tag = ReservedTags.PDF
        elif thumbnail_type == ThumbnailType.TEXT:
            tag = ReservedTags.TEXT
        else:
            raise ValueError(f"'{thumbnail_type}' is not a valid thumbnail type. Valid types are: {valid_types}")

        self._set_or_unset_thumbnails(tag, is_thumbnail)
        return self

    def _set_or_unset_thumbnails(self, type_tag, is_thumbnail):
        # Remove previous thumbnail type tags
        if self.schema.tags is not None:
            thumbnail_type_tags = {
                ReservedTags.VIDEO.value,
                ReservedTags.AUDIO.value,
                ReservedTags.IMAGE.value,
                ReservedTags.PDF.value,
                ReservedTags.TEXT.value,
            }

            for tag in thumbnail_type_tags:
                if tag in self.schema.tags:
                    self._remove_tag(tag)

        if is_thumbnail:
            self._add_tags({type_tag.value, ReservedTags.THUMBNAIL_VIZ.value})
        else:
            self._remove_tags(ReservedTags.THUMBNAIL_VIZ.value)

    def _set_or_unset(self, tag, is_set):
        if is_set:
            self._add_tags({tag})
        else:
            self._remove_tag(tag)

    def _add_tags(self, tags: Set[str]):
        if self.schema.tags is None:
            self.schema.tags = set()
        for t in tags:
            self.schema.tags.add(t)

    def _remove_tag(self, tag: str):
        if self.schema.tags is None:
            return
        try:
            self.schema.tags.remove(tag)
        except ValueError:
            logger.warning(f"Tag {tag} doesn't exist on the field, nothing to delete")

    def _remove_tags(self, *tags: str):
        if self.schema.tags is None:
            return
        for t in tags:
            self._remove_tag(t)

    @staticmethod
    def _get_backing_type(t: Union[Type, DagshubDataType]) -> MetadataFieldType:
        if issubclass(t, DagshubDataType):
            return t.backing_field_type

        if type(t) is type:
            if t not in metadataTypeLookup.keys():
                raise ValueError(f"Primitive type {type(t)} is not supported")
            return metadataTypeLookup[t]

        raise ValueError(f"{t} of type ({type(t)}) is not a valid primitive type or DagshubDataType")

    def apply(self):
        """
        Apply the outgoing changes to this metadata field.

        If you need to apply changes to multiple fields at once, use
        :func:`Datasource.apply_field_changes <dagshub.data_engine.model.datasource.Datasource.apply_field_changes>`
        instead.
        """
        self.datasource.apply_field_changes([self])



import logging
import mimetypes
import re
from dataclasses import dataclass, field
from pathlib import PurePosixPath, Path
from typing import Optional, Union, Mapping, Any, Dict, List
from os import PathLike
from dagshub.common.api.repo import RepoAPI, PathNotFoundError
from dagshub.data_engine.client.data_client import DataClient
from dagshub.data_engine.client.models import DatasourceType, DatasourceResult, PreprocessingStatus, MetadataFieldSchema
from dagshub.data_engine.model.datapoint import Datapoint
from dagshub.data_engine.model.errors import DatasourceAlreadyExistsError, DatasourceNotFoundError
from dagshub.common.util import multi_urljoin

from functools import cached_property

logger = logging.getLogger(__name__)

path_regexes = {
    DatasourceType.BUCKET: re.compile(r"(?P<schema>s3|gs|azure)://(?P<bucket>[\w\-._]+)(?P<prefix>/.*)?"),
    DatasourceType.REPOSITORY: re.compile(r"repo://(?P<user>[\w\-_.]+)/(?P<repo>[\w\-_.]+)(?P<prefix>/.*)?"),
}

expected_formats = {
    DatasourceType.BUCKET: "s3|gs|azure://bucket-name/prefix",
    DatasourceType.REPOSITORY: "repo://owner/reponame/prefix",
}


class InvalidPathFormatError(Exception):
    pass


@dataclass
class DatasourceState:
    repo: str
    name: Optional[str] = field(default=None)
    id: Optional[Union[int, str]] = field(default=None)

    source_type: DatasourceType = field(init=False)
    preprocessing_status: PreprocessingStatus = field(init=False)
    path: str = field(init=False)
    client: DataClient = field(init=False)
    repoApi: RepoAPI = field(init=False)
    metadata_fields: List[MetadataFieldSchema] = field(init=False)

    _revision: Optional[str] = field(init=False, default=None)

    def __post_init__(self):
        self.client = DataClient(self.repo)
        self.repoApi = RepoAPI(self.repo)
        if hasattr(self, "source_type") and self.source_type == DatasourceType.REPOSITORY:
            self.revision = self.path_parts()["revision"]
        if not hasattr(self, "metadata_fields"):
            self.metadata_fields = []

    @property
    def revision(self) -> str:
        """Used for repository sources, provides branch/revision from which to download files"""
        if self._revision is None:
            logger.warning("Revision wasn't set, assuming default repo branch")
            self.revision = self.repoApi.default_branch
        return self._revision

    @property
    def url(self) -> str:
        return multi_urljoin(self.repoApi.repo_url, f"datasets/datasource/{self.id}/gallery")

    @revision.setter
    def revision(self, val: str):
        self._revision = val

    def create(self):
        # Check that a datasource with this name doesn't exist
        datasources = self.client.get_datasources(self.id, self.name)
        if len(datasources) != 0:
            raise DatasourceAlreadyExistsError(self)

        create_result = self.client.create_datasource(self)
        self._update_from_ds_result(create_result)

    def get_from_dagshub(self):
        sources = self.client.get_datasources(self.id, self.name)
        if len(sources) == 0:
            raise DatasourceNotFoundError(self)
        elif len(sources) > 1:
            raise RuntimeError(
                f"Got too many ({len(sources)}) datasources with name '{self.name}' or id. Something went wrong"
            )
        self._update_from_ds_result(sources[0])

    def content_path(self, path: Union[str, Datapoint, Mapping[str, Any]]) -> str:
        """
        Returns the url for the content path of a specified path
        """
        path = self._extract_path(path)
        return multi_urljoin(self.root_content_path, path)

    def raw_path(self, path: Union[str, Datapoint, Mapping[str, Any]]) -> str:
        """
        Returns the url for the download path of a specified path
        """
        path = self._extract_path(path)
        return multi_urljoin(self.root_raw_path, path)

    @property
    def source_prefix(self) -> PurePosixPath:
        parts = self.path_parts()
        if parts.get("prefix"):
            return PurePosixPath(parts["prefix"].strip("/"))
        else:
            return PurePosixPath()

    def file_path(self, path: Union[str, Datapoint, Mapping[str, Any]]) -> PurePosixPath:
        """
        Returns the generic path of the path in the repo (adds prefix)
        """
        return self.source_prefix / self._extract_path(path)

    def blob_path(self, sha: str) -> str:
        """
        Returns the path for the blob of a datasource
        """
        return f"{self.repoApi.data_engine_url}/blob/{sha}"

    @cached_property
    def root_content_path(self) -> str:
        """
        Returns the root content path of the dataset for listing folders
        This is just a "prefix" of the datasource relative to the repo.
        In order to build a path of an entity you need to concatenate the path to this root
        """
        return self._root_path("content")

    @cached_property
    def root_raw_path(self):
        """
        Returns the root raw path of the dataset for downloading files
        This is just a "prefix" of the datasource relative to the repo.
        In order to build a path of an entity you need to concatenate the path to this root
        """
        return self._root_path("raw")

    def _root_path(self, path_type):
        assert path_type in ["raw", "content"]
        parts = self.path_parts()
        if self.source_type == DatasourceType.BUCKET:
            path_elems = [parts["schema"], parts["bucket"]]
            if parts.get("prefix"):
                path_elems.append(parts["prefix"])
            path_prefix = "/".join(path_elems)
            if path_type == "raw":
                return self.repoApi.storage_raw_api_url(path_prefix)
            elif path_type == "content":
                return self.repoApi.storage_content_api_url(path_prefix)
        elif self.source_type == DatasourceType.REPOSITORY:
            prefix = parts["prefix"]
            if prefix is None:
                prefix = ""
            # Assuming repo://user/repo is always the same user/repo we work with
            if path_type == "raw":
                return self.repoApi.raw_api_url(prefix, self.revision)
            elif path_type == "content":
                return self.repoApi.content_api_url(prefix, self.revision)
        elif self.source_type == DatasourceType.CUSTOM:
            raise NotImplementedError
        raise NotImplementedError

    def path_parts(self) -> Dict[str, str]:
        """
        Validates the provided path + returns a dictionary with elements that might be relevant for constructing paths
        """
        regex = path_regexes[self.source_type]
        match = regex.fullmatch(self.path)
        if match is None:
            raise InvalidPathFormatError(
                f"{self.path} is not valid path format for type {self.source_type}.\n"
                f"Expected format: {expected_formats[self.source_type]}"
            )
        res = match.groupdict()
        # For repository type - handle revision that is in format of repo://user/repo/branch:prefix
        # Couldn't do that with regexes, so handling it here
        if self.source_type == DatasourceType.REPOSITORY:
            res["revision"] = None
            if res.get("prefix"):
                prefix = res["prefix"]
                if ":" in prefix:
                    revision, prefix = prefix.split(":", 1)
                    res["revision"] = revision.strip("/")
                    if prefix.isspace():
                        prefix = None
                    elif not prefix.startswith("/"):
                        prefix = "/" + prefix
                    res["prefix"] = prefix
        return res

    @staticmethod
    def _extract_path(val: Union[str, Datapoint, Mapping[str, Any]]) -> str:
        if type(val) is str:
            return val
        elif type(val) is Datapoint:
            return val.path
        return val["path"]

    def _update_from_ds_result(self, ds: DatasourceResult):
        self.id = ds.id
        self.name = ds.name
        self.path = ds.rootUrl
        self.source_type = ds.type
        self.preprocessing_status = ds.preprocessingStatus
        self.metadata_fields = [] if ds.metadataFields is None else ds.metadataFields
        if self.source_type == DatasourceType.REPOSITORY:
            self.revision = self.path_parts()["revision"]

    @staticmethod
    def from_gql_result(repo: str, res: DatasourceResult):
        ds = DatasourceState(repo)
        ds._update_from_ds_result(res)
        return ds

    def import_metadata_from_file(self, datasource_name: str, file_path: Union[str, PathLike], path_column: str):
        load_location = self._determine_load_location(file_path)
        file_name = Path(file_path).name
        file_type = mimetypes.guess_type(file_path)[0] or "application/octet-stream"

        file_bytes = None
        if load_location == "disk":
            file_bytes = open(file_path, "rb")
        elif load_location == "repo":
            file_bytes = self.repoApi.get_file(file_path)

        files = {
            "file": (file_name, file_bytes, file_type),
        }
        data = {
            "datasource_name": datasource_name,
            "path_column": path_column,
        }

        url = multi_urljoin(self.repoApi.data_engine_url, "import/metadata")
        res = self.repoApi._http_request("POST", url, data=data, files=files)
        if res.status_code == 404:
            raise PathNotFoundError(f"Datasource {datasource_name} not found")
        elif res.status_code >= 400:
            error_msg = f"Got status code {res.status_code} when importing metadata to datasource {datasource_name}"
            logger.error(error_msg)
            logger.debug(res.content)
            raise RuntimeError(error_msg)
        return res.content

    def _determine_load_location(self, file_path: Union[str, Path]) -> str:
        # Local files take priority
        if Path(file_path).exists():
            return "disk"

        # Try to find it in the repo otherwise
        try:
            files = self.repoApi.list_path(Path(file_path).as_posix())
            if len(files) > 0:
                return "repo"
        except PathNotFoundError:
            pass

        # TODO: handle repo bucket too
        # TODO: improve and reuse https://github.com/DagsHub/client/pull/517#issuecomment-2288690281

        raise PathNotFoundError(f"Path {file_path} not found")



from typing import Literal

from dagshub.common.api.repo import RepoAPI
from dagshub.auth import get_token
from dagshub.common.helpers import log_message


def get_boto_client(repo_api: RepoAPI, token: str):
    """
    Creates a `boto3.client \
        <https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/boto3.html#boto3.client>`_.\
    object to interact with the bucket of the repository.

    `Read the boto3 docs for more <https://boto3.amazonaws.com/v1/documentation/api/latest/guide/s3-examples.html>`_

    :meta private:
    """
    endpoint_url = repo_api.repo_bucket_api_url()
    import boto3

    client = boto3.client("s3", endpoint_url=endpoint_url, aws_access_key_id=token, aws_secret_access_key=token)
    log_message(f"Client created. Use the name of the repo ({repo_api.repo_name}) as the name of the bucket")
    return client


def get_s3fs_client(repo_api: RepoAPI, token: str):
    """
    Creates an `s3fs.S3FileSystem <https://s3fs.readthedocs.io/en/latest/api.html#s3fs.core.S3FileSystem>`_\
    object to interact with the bucket of the repository

    `Read the s3fs docs for more <https://s3fs.readthedocs.io/en/latest/index.html#examples>`_

    :meta private:
    """
    endpoint_url = repo_api.repo_bucket_api_url()
    import s3fs

    client = s3fs.S3FileSystem(endpoint_url=endpoint_url, key=token, secret=token)
    log_message(f"Client created. Use the name of the repo ({repo_api.repo_name}) as the name of the bucket")
    return client


_s3_flavor_lookup = {
    "boto": get_boto_client,
    "s3fs": get_s3fs_client,
}

FlavorTypes = Literal["boto", "s3fs"]


def get_repo_bucket_client(repo: str, flavor: FlavorTypes = "boto"):
    """
    Creates an S3 client for the specified repository's DagsHub storage bucket

    Available flavors are:
        ``"boto"`` (Default): Returns a `boto3.client \
        <https://boto3.amazonaws.com/v1/documentation/api/latest/guide/s3-examples.html>`_.\
            with predefined EndpointURL and credentials.
            The name of the bucket is the name of the repository,
            and you will need to specify it for any request you make.

            Example usage::

                boto_client = get_repo_bucket_client("user/my-repo")

                # Upload file
                boto_client.upload_file(
                    Bucket="my-repo",      # name of the repo
                    Filename="local.csv",  # local path of file to upload
                    Key="remote.csv",      # remote path where to upload the file
                )
                # Download file
                boto_client.download_file(
                    Bucket="my-repo",      # name of the repo
                    Key="remote.csv",      # remote path from where to download the file
                    Filename="local.csv",  # local path where to download the file
                )


        ``"s3fs"``: Returns a \
            `s3fs.S3FileSystem <https://s3fs.readthedocs.io/en/latest/index.html#examples>`_\
             with predefined EndpointURL and credentials.
              The name of the bucket is the name of the repository,
              and you will need to specify it for any request you make

              Example usage::

                s3fs_client = get_repo_bucket_client("user/my-repo", flavor="s3fs")

                # Read from file
                with s3fs_client.open("my-repo/remote.csv", "rb") as f:
                    print(f.read())

                # Write to file
                with s3fs_client.open("my-repo/remote.csv", "wb") as f:
                    f.write(b"Content")

                # Upload file (can also upload directories)
                s3fs_client.put(
                    "local.csv",           # local path of file/dir to upload
                     "my-repo/remote.csv"  # remote path where to upload the file
                )



    Args:
        repo: Name of the repo in the format of ``username/reponame``
        flavor: one of the possible s3 client flavor variants
    """
    repo_api = RepoAPI(repo)
    token = get_token()
    return _s3_flavor_lookup[flavor](repo_api, token)



import logging

import appdirs
import os
from urllib.parse import urlparse
from dagshub import __version__
from httpx._client import USER_AGENT

logger = logging.getLogger(__name__)

HOST_KEY = "DAGSHUB_CLIENT_HOST"
DEFAULT_HOST = "https://dagshub.com"
CLIENT_ID_KEY = "DAGSHUB_CLIENT_ID"
DEFAULT_CLIENT_ID = "32b60ba385aa7cecf24046d8195a71c07dd345d9657977863b52e7748e0f0f28"
TOKENS_CACHE_LOCATION_KEY = "DAGSHUB_CLIENT_TOKENS_CACHE"
DEFAULT_TOKENS_CACHE_LOCATION = os.path.join(appdirs.user_cache_dir("dagshub"), "tokens")
TOKENS_CACHE_SCHEMA_VERSION = "1"
DAGSHUB_USER_TOKEN_KEY = "DAGSHUB_USER_TOKEN"
DAGSHUB_USERNAME_KEY = "DAGSHUB_USERNAME"
DAGSHUB_PASSWORD_KEY = "DAGSHUB_PASSWORD"
HTTP_TIMEOUT_KEY = "DAGSHUB_HTTP_TIMEOUT"
DAGSHUB_QUIET_KEY = "DAGSHUB_QUIET"


def set_host(new_host: str):
    _parsed_host = urlparse(new_host)
    _hostname = _parsed_host.hostname
    _host = _parsed_host.geturl().rstrip("/")

    global hostname, host, parsed_host
    hostname, host, parsed_host = _hostname, _host, _parsed_host


hostname = ""
host = ""
parsed_host = ""
set_host(os.environ.get(HOST_KEY, DEFAULT_HOST))

client_id = os.environ.get(CLIENT_ID_KEY, DEFAULT_CLIENT_ID)
cache_location = os.environ.get(TOKENS_CACHE_LOCATION_KEY, DEFAULT_TOKENS_CACHE_LOCATION)
token = os.environ.get(DAGSHUB_USER_TOKEN_KEY)
username = os.environ.get(DAGSHUB_USERNAME_KEY)
password = os.environ.get(DAGSHUB_PASSWORD_KEY)
custom_user_agent_suffix = f" dagshub-client-python/{__version__}"
requests_headers = {"user-agent": USER_AGENT + custom_user_agent_suffix}
http_timeout = os.environ.get(HTTP_TIMEOUT_KEY, 30)
REPO_INFO_URL = "api/v1/repos/{owner}/{reponame}"

quiet = bool(os.environ.get(DAGSHUB_QUIET_KEY, False))

# DVC config templates
CONFIG_GITIGNORE = "/config.local\n/tmp\n/cache"

RECOMMENDED_ANNOTATE_LIMIT_KEY = "RECOMMENDED_ANNOTATE_LIMIT"
recommended_annotate_limit = int(os.environ.get(RECOMMENDED_ANNOTATE_LIMIT_KEY, 1e5))

DATAENGINE_METADATA_UPLOAD_BATCH_SIZE_KEY = "DAGSHUB_DE_METADATA_UPLOAD_BATCH_SIZE"
dataengine_metadata_upload_batch_size = int(os.environ.get(DATAENGINE_METADATA_UPLOAD_BATCH_SIZE_KEY, 15000))

DISABLE_ANALYTICS_KEY = "DAGSHUB_DISABLE_ANALYTICS"
disable_analytics = "DAGSHUB_DISABLE_ANALYTICS" in os.environ

DOWNLOAD_THREADS_KEY = "DAGSHUB_DOWNLOAD_THREADS"
DEFAULT_DOWNLOAD_THREADS = 32
download_threads = int(os.environ.get(DOWNLOAD_THREADS_KEY, DEFAULT_DOWNLOAD_THREADS))

UPLOAD_THREADS_KEY = "DAGSHUB_UPLOAD_THREADS"
DEFAULT_UPLOAD_THREADS = 8
upload_threads = int(os.environ.get(UPLOAD_THREADS_KEY, DEFAULT_UPLOAD_THREADS))

if download_threads > DEFAULT_DOWNLOAD_THREADS:
    logger.warning(
        f"Number of download threads was set to {download_threads}. "
        f"We recommend lowering the value if you get met with rate limits"
    )



import base64
import datetime
import json
import logging
import tempfile
import math
import os.path
import threading
import time
import uuid
import requests
import webbrowser
from contextlib import contextmanager
from dataclasses import dataclass, field
from os import PathLike
from pathlib import Path
from typing import Any, Dict, List, Optional, TYPE_CHECKING, Union, Set, ContextManager, Tuple, Literal, Callable
from tenacity import retry, wait_fixed, stop_after_attempt, retry_if_exception_type


import rich.progress
from dataclasses_json import config, LetterCase, DataClassJsonMixin
from pathvalidate import sanitize_filepath

import dagshub.common.config
from dagshub.common import rich_console
from dagshub.common.analytics import send_analytics_event
from dagshub.common.environment import is_mlflow_installed
from dagshub.common.helpers import prompt_user, http_request, log_message
from dagshub.common.rich_util import get_rich_progress
from dagshub.common.util import lazy_load, multi_urljoin, to_timestamp, exclude_if_none, deprecated
from dagshub.data_engine.annotation.importer import AnnotationImporter, AnnotationType, AnnotationLocation
from dagshub.data_engine.client.models import (
    PreprocessingStatus,
    MetadataFieldSchema,
    ScanOption,
    autogenerated_columns,
    DatasetResult,
)
from dagshub.data_engine.dtypes import MetadataFieldType
from dagshub.data_engine.model.datapoint import Datapoint
from dagshub.data_engine.model.errors import (
    WrongOperatorError,
    WrongOrderError,
    DatasetFieldComparisonError,
    FieldNotFoundError,
    DatasetNotFoundError,
    LSInitializingError,
)
from dagshub.data_engine.model.metadata import (
    validate_uploading_metadata,
    run_preupload_transforms,
    precalculate_metadata_info,
)
from dagshub.data_engine.model.metadata import wrap_bytes
from dagshub.data_engine.model.metadata_field_builder import MetadataFieldBuilder
from dagshub.data_engine.model.query import QueryFilterTree
from dagshub.data_engine.model.schema_util import metadataTypeLookup, metadataTypeLookupReverse
from dagshub.data_engine.model.datasource_state import DatasourceState

if TYPE_CHECKING:
    from dagshub.data_engine.model.query_result import QueryResult
    import fiftyone as fo
    import pandas
    import mlflow
    import mlflow.entities
    import cloudpickle
    import ngrok
else:
    plugin_server_module = lazy_load("dagshub.data_engine.voxel_plugin_server.server")
    fo = lazy_load("fiftyone")
    mlflow = lazy_load("mlflow")
    pandas = lazy_load("pandas")
    ngrok = lazy_load("ngrok")
    cloudpickle = lazy_load("cloudpickle")

logger = logging.getLogger(__name__)

LS_ORCHESTRATOR_URL = "http://127.0.0.1"
MLFLOW_DATASOURCE_TAG_NAME = "dagshub.datasets.datasource_id"
MLFLOW_DATASET_TAG_NAME = "dagshub.datasets.dataset_id"


@dataclass
class DatapointMetadataUpdateEntry(DataClassJsonMixin):
    url: str
    key: str
    value: str
    valueType: MetadataFieldType = field(metadata=config(encoder=lambda val: val.value))
    allowMultiple: bool = False
    timeZone: Optional[str] = field(
        default=None, metadata=config(exclude=exclude_if_none, letter_case=LetterCase.CAMEL)
    )


@dataclass
class DatapointDeleteMetadataEntry(DataClassJsonMixin):
    datapointId: str
    key: str


@dataclass
class DatapointDeleteEntry(DataClassJsonMixin):
    datapointId: str


@dataclass
class Field:
    """
    Class used to define custom fields for use in \
    :func:`Datasource.select() <dagshub.data_engine.model.datasource.Datasource.select>` or in filtering.

    Example of filtering on old data from a field::

        t = datetime.now() - timedelta(days=2)
        q = ds[Field("size", as_of=t)] > 500
        q.all()
    """

    field_name: str
    """The database field where the values are stored. In other words, where to get the values from."""
    as_of: Optional[Union[float, datetime.datetime]] = None
    """
    If defined, the data in this field would be shown as of this moment in time.

    Accepts either a datetime object, or a UTC timestamp.
    """
    alias: Optional[str] = None
    """
    How the returned custom data field should be named.

    Useful when you're comparing the same field at multiple points in time::

        yesterday = datetime.now() - timedelta(days=1)

        ds.select(
            Field("value", alias="value_today"),
            Field("value", as_of=yesterday, alias="value_yesterday")
        ).all()
    """

    @property
    def as_of_timestamp(self) -> Optional[int]:
        if self.as_of is None:
            return None
        return to_timestamp(self.as_of)

    def to_dict(self, ds: "Datasource") -> Dict[str, Any]:
        if not ds.has_field(self.field_name):
            raise FieldNotFoundError(self.field_name)

        res_dict: Dict[str, Union[str, int, None]] = {"name": self.field_name}
        if self.as_of is not None:
            res_dict["asOf"] = self.as_of_timestamp
        if self.alias:
            res_dict["alias"] = self.alias
        return res_dict


_metadata_contexts: Dict[Union[int, str], "MetadataContextManager"] = {}


class Datasource:
    def __init__(
        self,
        datasource: "DatasourceState",
        query: Optional["DatasourceQuery"] = None,
        from_dataset: Optional["DatasetState"] = None,
    ):
        self._source = datasource
        if query is None:
            query = DatasourceQuery()
        self._query = query
        # this ref marks if source is currently used in
        # meta-data update 'with' block
        self._explicit_update_ctx: Optional[MetadataContextManager] = None
        self.assigned_dataset = from_dataset

        self.ngrok_listener = None

    @property
    def has_explicit_context(self):
        return self._explicit_update_ctx is not None

    @property
    def source(self) -> "DatasourceState":
        return self._source

    def clear_query(self, reset_to_dataset=True):
        """
        Clear the attached query.

        Args:
            reset_to_dataset: If ``True`` and this Datasource was saved as a dataset, reset to the query in the dataset,
                instead of clearing the query completely.
        """
        if reset_to_dataset and self.assigned_dataset is not None and self.assigned_dataset.query is not None:
            self._query = self.assigned_dataset.query.__deepcopy__()
        else:
            self._query = DatasourceQuery()

    def __deepcopy__(self, memodict={}) -> "Datasource":
        res = Datasource(self._source, self._query.__deepcopy__())
        res.assigned_dataset = self.assigned_dataset

        return res

    def get_query(self) -> "DatasourceQuery":
        return self._query

    @property
    def annotation_fields(self) -> List[str]:
        """Return all fields that have the annotation meta tag set"""
        return [f.name for f in self.fields if f.is_annotation()]

    @property
    def document_fields(self) -> List[str]:
        return [f.name for f in self.fields if f.is_document()]

    def serialize_gql_query_input(self) -> Dict:
        """
        Serialize the query of this Datasource for use in GraphQL querying (e.g. getting datapoints)

        :meta private:
        """
        return self._query.to_dict()

    def _deserialize_from_gql_result(self, query_dict: Dict):
        """
        Imports query information from ``query_dict``
        """
        self._query = DatasourceQuery.from_dict(query_dict)

    def load_from_dataset(
        self, dataset_id: Optional[Union[str, int]] = None, dataset_name: Optional[str] = None, change_query=True
    ):
        """
        Imports query information from a dataset with the specified id or name. Either of id or name could be specified

        Args:
            dataset_id: ID of the dataset
            dataset_name: Name of the dataset
            change_query: Whether to change the query of this object to the query in the dataset

        :meta private:
        """
        datasets = self.source.client.get_datasets(id=dataset_id, name=dataset_name)
        if not datasets:
            raise DatasetNotFoundError(self.source.repo, dataset_id, dataset_name)
        dataset_state = DatasetState.from_gql_dataset_result(datasets[0])
        self.load_from_dataset_state(dataset_state, change_query)

    def load_from_dataset_state(self, dataset_state: "DatasetState", change_query=True):
        """
        Imports query information from a :class:`~dagshub.data_engine.model.dataset_state.DatasetState`

        Args:
            dataset_state: State to load
            change_query: If false, only assigns the dataset.
                If true, also changes the query to be the query of the dataset

        :meta private:
        """
        if self.source.id != dataset_state.datasource_id:
            raise RuntimeError(
                "Dataset belongs to a different datasource "
                f"(This datasource: {self.source.id}, Dataset's datasource: {dataset_state.datasource_id})"
            )
        self.assigned_dataset = dataset_state
        if change_query:
            self._query = dataset_state.query

    def sample(self, start: Optional[int] = None, end: Optional[int] = None) -> "QueryResult":
        if start is not None:
            logger.warning("Starting slices is not implemented for now")
        res = self._source.client.sample(self, end, include_metadata=True)
        res._load_autoload_fields()
        return res

    def head(self, size=100, load_documents=True, load_annotations=True) -> "QueryResult":
        """
        Executes the query and returns a :class:`.QueryResult` object containing first ``size`` datapoints

        Args:
            size: how many datapoints to get. Default is 100
            load_documents: Automatically download all document blob fields
            load_annotations: Automatically download all annotation blob fields
        """
        self._check_preprocess()
        send_analytics_event("Client_DataEngine_DisplayTopResults", repo=self.source.repoApi)
        res = self._source.client.head(self, size)
        res._load_autoload_fields(documents=load_documents, annotations=load_annotations)
        return res

    def all(self, load_documents=True, load_annotations=True) -> "QueryResult":
        """
        Executes the query and returns a :class:`.QueryResult` object containing all datapoints

        If there's an active MLflow run, logs an artifact with information about the query to the run.

        Args:
            load_documents: Automatically download all document blob fields
            load_annotations: Automatically download all annotation blob fields
        """
        self._check_preprocess()
        res = self._source.client.get_datapoints(self)
        self._autolog_mlflow(res)
        res._load_autoload_fields(documents=load_documents, annotations=load_annotations)

        return res

    def select(self, *selected: Union[str, Field]) -> "Datasource":
        """
        Select which fields should appear in the query result.

        If you want to query older versions of metadata,
        use :class:`Field` objects with ``as_of`` set to your desired time.

        By default, only the defined fields are returned.
        If you want to return all existing fields plus whatever additional fields you define,
        add ``"*"`` into the arguments.

        Args:
            selected: Fields you want to select. Can be either of:

                - Name of the field to select: ``"field"``.
                - ``"*"`` to select all the fields in the datasource.
                - :class:`Field` object.

        Example::

            t = datetime.now() - timedelta(hours=24)
            q1 = ds.select("*", Field("size", as_of=t, alias="size_asof_24h_ago"))
            q1.all()
        """

        include_all = False
        selects = []
        for s in selected:
            if isinstance(s, Field):
                selects.append(s.to_dict(self))
            else:
                if s != "*":
                    selects.append({"name": s})
                else:
                    include_all = True

        if include_all:
            aliases = set([s["alias"] for s in selects if "alias" in s])
            for f in self.fields:
                if f.name in aliases:
                    raise ValueError(
                        f"Alias {f.name} can't be used, because * was specified and "
                        f"a field with that name already exists"
                    )
                selects.append({"name": f.name})

        new_ds = self.__deepcopy__()
        new_ds.get_query().select = selects
        return new_ds

    def as_of(self, time: Union[float, datetime.datetime]) -> "Datasource":
        """
        Get a snapshot of the datasource's state as of ``time``.

        Args:
            time: At which point in time do you want to get data from.\
                Either a UTC timestamp or a ``datetime`` object.

        In the following example, you will get back datapoints that were created no later than yesterday AND \
        had their size at this point bigger than 5 bytes::

            t = datetime.now() - timedelta(hours=24)
            q1 = (ds["size"] > 5).as_of(t)
            q1.all()

        .. note::
            If used with :func:`select`, the ``as_of`` set on the fields takes precedence
            over the global query ``as_of`` set here.
        """
        new_ds = self.__deepcopy__()

        new_ds._query.as_of = to_timestamp(time)
        return new_ds

    def with_time_zone(self, tz_val: str) -> "Datasource":
        """
        A time zone offset string in the form of "+HH:mm" or "-HH:mm".

        A metadata of type datetime is always stored in DB as a UTC time, when a query is done on this field
        there are 3 options:

        - Metadata was saved with a timezone, in which case it will be used.

        - Metadata was saved without a timezone, in which case UTC will be used.

        - with_time_zone specified a time zone and it will override whatever is in the database.
        """
        new_ds = self.__deepcopy__()

        new_ds._query.time_zone = tz_val
        return new_ds

    def order_by(self, *args: Union[str, Tuple[str, Union[bool, str]]]) -> "Datasource":
        """
        Sort the query result by the specified fields.
        Any previously set order will be overwritten.

        Args:
            Fields to sort by. Can be either of:
                - Name of the field to sort by: ``"field"``.
                - A tuple of ``(field_name, ascending)``: ``("field", True)``.
                - A tuple of ``(field_name, "asc"|"desc")``: ``("field", "asc")``.

        Examples::

            ds.order_by("size").all()                   # Order by ascending size
            ds.order_by(("date", "desc"), "size).all()  # Order by descending date, then ascending size
        """
        new_ds = self.__deepcopy__()
        orders = []
        for arg in args:
            if isinstance(arg, str):
                orders.append({"field": arg, "order": "ASC"})
            else:
                if len(arg) != 2:
                    raise RuntimeError(
                        f"Invalid sort argument {arg}, must be a tuple (<field>, 'asc'|'desc'|True|False)"
                    )
                if isinstance(arg[1], bool):
                    order = "ASC" if arg[1] else "DESC"
                elif isinstance(arg[1], str) and arg[1].upper() in ["ASC", "DESC"]:
                    order = arg[1].upper()
                else:
                    raise RuntimeError(f"Invalid sort argument {arg}, second value must be 'asc'|'desc'|True|False")
                orders.append({"field": arg[0], "order": order})
        new_ds.get_query().order_by = orders
        return new_ds

    def _check_preprocess(self):
        self.source.get_from_dagshub()
        if (
            self.source.preprocessing_status == PreprocessingStatus.IN_PROGRESS
            or self.source.preprocessing_status == PreprocessingStatus.AUTO_SCAN_IN_PROGRESS
        ):
            logger.warning(
                f"Datasource {self.source.name} is currently in the progress of rescanning. "
                f"Values might change if you requery later"
            )

    def metadata_field(self, field_name: str) -> MetadataFieldBuilder:
        """
        Returns a builder for a metadata field.
        The builder can be used to change properties of a field or create a new field altogether.
        Note that fields get automatically created when you upload new metadata to the Data Engine,
        so it's not necessary to create fields with this function.

        Example of creating a new annotation field::

            ds.metadata_field("annotation").set_type(dtypes.LabelStudioAnnotation).apply()

        .. note::
            New fields have to have their type defined using ``.set_type()`` before doing anything else

        Example of marking an existing field as an annotation field::

            ds.metadata_field("existing-field").set_annotation().apply()

        Args:
            field_name: Name of the field that you want to create/change

        """
        return MetadataFieldBuilder(self, field_name)

    def apply_field_changes(self, field_builders: List[MetadataFieldBuilder]):
        """
        Applies one or multiple metadata field builders
        that can be constructed using the :func:`metadata_field()` function.
        """
        self.source.client.update_metadata_fields(self, [builder.schema for builder in field_builders])
        self.source.get_from_dagshub()

    @property
    def implicit_update_context(self) -> "MetadataContextManager":
        """
        Context that is used when updating metadata through ``dp[field] = value`` syntax, can be created on demand.

        :meta private:
        """
        key = self.source.id
        if key not in _metadata_contexts:
            _metadata_contexts[key] = MetadataContextManager(self)
        return _metadata_contexts[key]

    def upload_metadata_of_implicit_context(self):
        """
        commit meta data changes done in dictionary assignment context
        :meta private:
        """
        try:
            self._upload_metadata(self.implicit_update_context.get_metadata_entries())
        finally:
            self.implicit_update_context.clear()

    def metadata_context(self) -> ContextManager["MetadataContextManager"]:
        """
        Returns a metadata context, that you can upload metadata through using its
        :func:`~MetadataContextManager.update_metadata()` function.
        Once the context is exited, all metadata is uploaded in one batch::

            with ds.metadata_context() as ctx:
                ctx.update_metadata("file1", {"key1": True, "key2": "value"})

        """

        # Need to have the context manager inside a wrapper to satisfy MyPy + PyCharm type hinter
        @contextmanager
        def func():
            self.source.get_from_dagshub()
            send_analytics_event("Client_DataEngine_addEnrichments", repo=self.source.repoApi)
            ctx = MetadataContextManager(self)

            self._explicit_update_ctx = ctx
            yield ctx
            try:
                entries = ctx.get_metadata_entries() + self.implicit_update_context.get_metadata_entries()
                self._upload_metadata(entries)
            finally:
                # Clear the implicit context because it can persist
                self.implicit_update_context.clear()
                # The explicit one created with with: can go away
                self._explicit_update_ctx = None

        return func()

    def upload_metadata_from_file(
        self, file_path, path_column: Optional[Union[str, int]] = None, ingest_on_server: bool = False
    ):
        """
        Upload metadata from a file.

        Args:
            file_path: Path to the file with metadata. Allowed formats are CSV, Parquet, ZIP, GZ.
            path_column: Column with the datapoints' paths. Can either be the name of the column, or its index.
                If not specified, the first column is used.
            ingest_on_server: Set to ``True`` to process the metadata asynchronously.
                The file will be sent to our server and ingested into the datasource there.
                Default is ``False``.
        """
        send_analytics_event("Client_DataEngine_addEnrichmentsWithFile", repo=self.source.repoApi)

        if ingest_on_server:
            datasource_name = self.source.name
            self._source.import_metadata_from_file(datasource_name, file_path, path_column)
        else:
            df = self._convert_file_to_df(file_path)
            self.upload_metadata_from_dataframe(df, path_column, ingest_on_server)

    def upload_metadata_from_dataframe(
        self, df: "pandas.DataFrame", path_column: Optional[Union[str, int]] = None, ingest_on_server: bool = False
    ):
        """
        Upload metadata from a pandas dataframe.

        All columns are uploaded as metadata, and the path of every datapoint is taken from `path_column`.

        Args:
            df (`pandas.DataFrame <https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.html>`_):
                DataFrame with metadata
            path_column: Column with the datapoints' paths. Can either be the name of the column, or its index.
                If not specified, the first column is used.
            ingest_on_server: Set to ``True`` to process the metadata asynchronously.
                The file will be sent to our server and ingested into the datasource there.
                Default is ``False``.
        """
        self.source.get_from_dagshub()
        send_analytics_event("Client_DataEngine_addEnrichmentsWithDataFrame", repo=self.source.repoApi)

        if ingest_on_server:
            self._remote_upload_metadata_from_dataframe(df, path_column)
        else:
            metadata = self._df_to_metadata(df, path_column, multivalue_fields=self._get_multivalue_fields())
            self._upload_metadata(metadata)

    def _remote_upload_metadata_from_dataframe(
        self, df: "pandas.DataFrame", path_column: Optional[Union[str, int]] = None
    ):
        datasource_name = self.source.name

        with tempfile.NamedTemporaryFile(suffix=".parquet", delete=True) as tmp:
            file_path = tmp.name
            df.to_parquet(file_path, index=False)

            self._source.import_metadata_from_file(datasource_name, file_path, path_column)

    def _get_multivalue_fields(self) -> Set[str]:
        res = set()
        for col in self.source.metadata_fields:
            if col.multiple:
                res.add(col.name)
        return res

    def _df_to_metadata(
        self, df: "pandas.DataFrame", path_column: Optional[Union[str, int]] = None, multivalue_fields=set()
    ) -> List[DatapointMetadataUpdateEntry]:
        res: List[DatapointMetadataUpdateEntry] = []
        if path_column is None:
            path_column = df.columns[0]
        elif isinstance(path_column, str):
            if path_column not in df.columns:
                raise RuntimeError(f"Column {path_column} does not exist in the dataframe")
        elif isinstance(path_column, int):
            path_column = df.columns[path_column]

        # objects are actually mixed and not guaranteed to be string, but this should cover most use cases
        if df.dtypes[path_column] != "object":
            raise RuntimeError(f"Column {path_column} doesn't have strings")

        field_value_types = {f.name: f.valueType for f in self.fields}
        document_fields = set(self.document_fields)

        for _, row in df.iterrows():
            datapoint = row[path_column]
            for key, val in row.items():
                if key == path_column:
                    continue
                key = str(key)
                if key in autogenerated_columns:
                    continue
                if val is None:
                    continue
                # Pandas specific: since pandas doesn't distinguish between None and NaN, don't upload it
                if isinstance(val, float) and math.isnan(val):
                    continue
                # Pandas specific: Cast pandas Timestamp to datetime
                if isinstance(val, list) or isinstance(val, pandas._libs.tslibs.timestamps.Timestamp):
                    if isinstance(val, list):
                        if isinstance(val[0], pandas._libs.tslibs.timestamps.Timestamp):
                            val = [val.to_pydatetime() for val in val]
                    else:
                        val = val.to_pydatetime()
                if isinstance(val, list):
                    if key not in multivalue_fields:
                        multivalue_fields.add(key)
                        # Promote all the existing uploading metadata to multivalue
                        for update_entry in res:
                            if update_entry.key == key:
                                update_entry.allowMultiple = True
                    for sub_val in val:
                        value_type = field_value_types.get(key)
                        time_zone = None
                        if value_type is None:
                            value_type = metadataTypeLookup[type(sub_val)]
                            field_value_types[key] = value_type
                        # Don't override bytes if they're not bytes - probably just undownloaded values
                        if value_type == MetadataFieldType.BLOB and not isinstance(sub_val, bytes):
                            if key not in document_fields:
                                continue
                        # Pandas quirk - integers are floats on the backend
                        if value_type == MetadataFieldType.INTEGER:
                            sub_val = int(sub_val)
                        if isinstance(sub_val, str) and key in document_fields:
                            sub_val = sub_val.encode("utf-8")
                        if isinstance(sub_val, bytes):
                            sub_val = wrap_bytes(sub_val)
                        if isinstance(sub_val, datetime.datetime):
                            time_zone = _get_datetime_utc_offset(sub_val)
                            sub_val = int(sub_val.timestamp() * 1000)
                        res.append(
                            DatapointMetadataUpdateEntry(
                                url=datapoint,
                                key=key,
                                value=str(sub_val),
                                valueType=value_type,
                                allowMultiple=True,
                                timeZone=time_zone,
                            )
                        )
                else:
                    value_type = field_value_types.get(key)
                    if value_type is None:
                        value_type = metadataTypeLookup[type(val)]
                        field_value_types[key] = value_type
                    # Don't override bytes if they're not bytes - probably just undownloaded values
                    if value_type == MetadataFieldType.BLOB and not isinstance(val, bytes):
                        if key not in document_fields:
                            continue
                    time_zone = None
                    # Pandas quirk - integers are floats on the backend
                    if value_type == MetadataFieldType.INTEGER:
                        val = int(val)
                    if isinstance(val, str) and key in document_fields:
                        val = val.encode("utf-8")
                    if isinstance(val, bytes):
                        val = wrap_bytes(val)
                    if isinstance(val, datetime.datetime):
                        time_zone = _get_datetime_utc_offset(val)
                        val = int(val.timestamp() * 1000)
                    res.append(
                        DatapointMetadataUpdateEntry(
                            url=datapoint,
                            key=key,
                            value=str(val),
                            valueType=value_type,
                            allowMultiple=key in multivalue_fields,
                            timeZone=time_zone,
                        )
                    )
        return res

    def delete_source(self, force: bool = False):
        """
        Delete the record of this datasource along with all datapoints.

        .. warning::
            This is a destructive operation! If you delete the datasource,
            all the datapoints and metadata will be removed.

        Args:
            force: Skip the confirmation prompt
        """
        prompt = (
            f'You are about to delete datasource "{self.source.name}" for repo "{self.source.repo}"\n'
            f"This will remove the datasource and ALL datapoints "
            f"and metadata records associated with the source."
        )
        if not force:
            user_response = prompt_user(prompt)
            if not user_response:
                print("Deletion cancelled")
                return
        self.source.client.delete_datasource(self)

    def delete_dataset(self, force: bool = False):
        """
        Deletes the dataset, if this object was created from a dataset
        (e.g. from :func:`.datasets.get_dataset()`).

        This doesn't delete the underlying datasource and its metadata, only deleting the dataset and its query.

        If this datasource object wasn't created from a dataset, raises a ``ValueError``.

        Args:
            force: Skip the confirmation prompt
        """
        if self.assigned_dataset is None:
            raise ValueError("This datasource was not created from a dataset")
        prompt = (
            f'You are about to delete dataset "{self.assigned_dataset.dataset_name}" for repo "{self.source.repo}"\n'
            f'The datasource "{self.source.name}" will still exist, but the dataset entry will be removed'
        )
        if not force:
            user_response = prompt_user(prompt)
            if not user_response:
                print("Deletion cancelled")
                return
        self.source.client.delete_dataset(self.assigned_dataset.dataset_id)

    def scan_source(self, options: Optional[List[ScanOption]] = None):
        """
        This function fires a call to the backend to rescan the datapoints.
        Call this function whenever you uploaded new files and want them to appear when querying the datasource,
        or if you changed existing file contents and want their metadata to be updated.

        DagsHub periodically rescans all datasources, this function is a way to make a scan happen as soon as possible.

        Notes about automatically scanned metadata:
            1. Only new datapoints (files) will be added.
               If files were removed from the source, their metadata will still remain,
               and they will still be returned from queries on the datasource.
               An API to actively remove metadata will be available soon.
            2. Some metadata fields will be automatically scanned and updated by DagsHub based on this scan -
               the list of automatic metadata fields is growing frequently!

        Args:
            options: List of scanning options. If not sure, leave empty.
        """
        logger.debug("Rescanning datasource")
        self.source.client.scan_datasource(self, options=options)

    def _upload_metadata(self, metadata_entries: List[DatapointMetadataUpdateEntry]):
        precalculated_info = precalculate_metadata_info(self, metadata_entries)
        validate_uploading_metadata(precalculated_info)
        run_preupload_transforms(self, metadata_entries, precalculated_info)

        progress = get_rich_progress(rich.progress.MofNCompleteColumn())

        upload_batch_size = dagshub.common.config.dataengine_metadata_upload_batch_size
        total_entries = len(metadata_entries)
        total_task = progress.add_task(f"Uploading metadata (batch size {upload_batch_size})...", total=total_entries)

        with progress:
            for start in range(0, total_entries, upload_batch_size):
                entries = metadata_entries[start : start + upload_batch_size]
                logger.debug(f"Uploading {len(entries)} metadata entries...")
                self.source.client.update_metadata(self, entries)
                progress.update(total_task, advance=upload_batch_size)
            progress.update(total_task, completed=total_entries, refresh=True)

        # Update the status from dagshub, so we get back the new metadata columns
        self.source.get_from_dagshub()

    def delete_metadata_from_datapoints(self, datapoints: List[Datapoint], fields: List[str]):
        """
        Delete metadata from datapoints.
        The deleted values can be accessed using versioned query with time set before the deletion

        Args:
            datapoints: datapoints to delete metadata from
            fields: fields to delete
        """
        metadata_entries = []
        for d in datapoints:
            for n in fields:
                metadata_entries.append(DatapointDeleteMetadataEntry(datapointId=d.datapoint_id, key=n))
        self.source.client.delete_metadata_for_datapoint(self, metadata_entries)

    def delete_datapoints(self, datapoints: List[Datapoint], force: bool = False):
        """
        Delete datapoints.

        - These datapoints will no longer show up in queries.
        - Does not delete the datapoint's file, only removing the data from the datasource.
        - You can still query these datapoints and associated metadata with \
        versioned queries whose time is before deletion time.
        - You can re-add these datapoints to the datasource by uploading new metadata to it with, for example, \
        :func:`Datasource.metadata_context <dagshub.data_engine.model.datasource.Datasource.metadata_context>`. \
        This will create a new datapoint with new id and new metadata records.
        - Datasource scanning will *not* add these datapoints back.

        Args:
            datapoints: list of datapoints objects to delete
            force: Skip the confirmation prompt
        """
        dps_str = "\n\t".join([""] + [d.path for d in datapoints])
        prompt = (
            f"You are about to delete the following datapoint(s): {dps_str}\n"
            f"This will remove the datapoint and metadata from unversioned queries, "
            f"but won't delete the underlying file."
        )
        if not force:
            user_response = prompt_user(prompt)
            if not user_response:
                print("Deletion cancelled")
                return

        self.source.client.delete_datapoints(
            self, [DatapointDeleteEntry(datapointId=d.datapoint_id) for d in datapoints]
        )

    def save_dataset(self, name: str) -> "Datasource":
        """
        Save the dataset, which is a combination of datasource + query, on the backend.
        That way you can persist and share your queries.
        You can get the dataset back later by calling :func:`.datasets.get_dataset()`

        Args:
            name: Name of the dataset

        Returns:
            A datasource object with the dataset assigned to it
        """
        send_analytics_event("Client_DataEngine_QuerySaved", repo=self.source.repoApi)

        self.source.client.save_dataset(self, name)
        log_message(f"Dataset {name} saved")

        copy_with_ds_assigned = self.__deepcopy__()
        copy_with_ds_assigned.load_from_dataset(dataset_name=name, change_query=False)
        return copy_with_ds_assigned

    @deprecated("Either use autologging, or QueryResult.log_to_mlflow() if autologging is turned off")
    def log_to_mlflow(
        self,
        artifact_name: Optional[str] = None,
        run: Optional["mlflow.entities.Run"] = None,
        as_of: Optional[datetime.datetime] = None,
    ) -> "mlflow.Entities.Run":
        """
        Logs the current datasource state to MLflow as an artifact.

        .. warning::
            This function is deprecated. Use autologging or
            :func:`QueryResult.log_to_mlflow() <dagshub.data_engine.model.query_result.QueryResult.log_to_mlflow>`
            instead.

        Args:
            artifact_name: Name of the artifact that will be stored in the MLflow run.
            run: MLflow run to save to. If ``None``, uses the active MLflow run or creates a new run.
            as_of: The querying time for which to save the artifact.
                Any time the datasource is recreated from the artifact, it will be queried as of this timestamp.
                If None, the current machine time will be used.
                If the artifact is autologged to MLflow (will happen if you have an active MLflow run),
                then the timestamp of the query will be used.

        Returns:
            Run to which the artifact was logged.
        """
        if artifact_name is None:
            as_of = as_of or (self._query.as_of or datetime.datetime.now())
            artifact_name = self._get_mlflow_artifact_name("log", as_of)
        elif not artifact_name.endswith(".dagshub.dataset.json"):
            artifact_name += ".dagshub.dataset.json"

        return self._log_to_mlflow(artifact_name, run, as_of)

    def _autolog_mlflow(self, qr: "QueryResult"):
        if not is_mlflow_installed:
            return
        # Run ONLY if there's an active run going on
        active_run = mlflow.active_run()
        if active_run is None:
            return

        artifact_name = self._get_mlflow_artifact_name("autolog", qr.query_data_time)

        threading.Thread(
            target=self._log_to_mlflow,
            kwargs={"artifact_name": artifact_name, "run": active_run, "as_of": qr.query_data_time},
        ).start()

    def _log_to_mlflow(
        self,
        artifact_name,
        run: Optional["mlflow.entities.Run"] = None,
        as_of: Optional[datetime.datetime] = None,
    ) -> "mlflow.Entities.Run":
        if run is None:
            run = mlflow.active_run()
            if run is None:
                run = mlflow.start_run()
        client = mlflow.MlflowClient()
        client.set_tag(run.info.run_id, MLFLOW_DATASOURCE_TAG_NAME, self.source.id)
        if self.assigned_dataset is not None:
            client.set_tag(run.info.run_id, MLFLOW_DATASET_TAG_NAME, self.assigned_dataset.dataset_id)
        client.log_dict(run.info.run_id, self._to_dict(as_of), artifact_name)
        log_message(f'Saved the datasource state to MLflow (run "{run.info.run_name}") as "{artifact_name}"')
        return run

    def _get_mlflow_artifact_name(self, prefix: str, as_of: datetime.datetime) -> str:
        now_time = as_of.strftime("%Y-%m-%dT%H-%M-%S")  # Not ISO format to make it a valid filename
        uuid_chunk = str(uuid.uuid4())[-4:]
        return f"{prefix}_{self.source.name}_{now_time}_{uuid_chunk}.dagshub.dataset.json"

    def save_to_file(self, path: Union[str, PathLike] = ".") -> Path:
        """
        Saves a JSON file representing the current state of datasource or dataset.
        Useful for connecting code versions to the datasource used for training.

        .. note::
            Does not save the actual contents of the datasource/dataset, only the query.

        Args:
            path: Where to save the file. If path is an existing folder, saves to ``<path>/<ds_name>.json``.

        Returns:
            The path to the saved file
        """
        path = Path(path)
        if path.is_dir():
            if self.assigned_dataset is not None:
                name = self.assigned_dataset.dataset_name
            else:
                name = self.source.name
            path = path / f"{name}.json"
        path.parent.mkdir(parents=True, exist_ok=True)
        res = self._to_dict()
        with open(path, "w") as file:
            json.dump(res, file, indent=4, sort_keys=True)
        log_message(f"Datasource saved to '{path}'")

        return path

    def _serialize(self, as_of: datetime.datetime) -> "DatasourceSerializedState":
        res = DatasourceSerializedState(
            repo=self.source.repo,
            datasource_id=self.source.id,
            datasource_name=self.source.name,
            query=self._query,
            timestamp=as_of.timestamp(),
            modified=self.is_query_different_from_dataset,
            link=self._generate_visualize_url(),
        )
        if self.assigned_dataset is not None:
            res.dataset_id = self.assigned_dataset.dataset_id
            res.dataset_name = self.assigned_dataset.dataset_name
        if self._query.as_of is not None:
            res.timed_link = res.link
        elif as_of is not None:
            timed_ds = self.as_of(as_of)
            res.timed_link = timed_ds._generate_visualize_url()
        return res

    def _to_dict(self, as_of: Optional[datetime.datetime] = None) -> Dict:
        if as_of is None:
            as_of = datetime.datetime.now()
        res = self._serialize(as_of).to_dict()
        # Skip Nones in the result
        res = {k: v for k, v in res.items() if v is not None}
        return res

    @property
    def is_query_different_from_dataset(self) -> Optional[bool]:
        """
        Is the current query of the object different from the one in the assigned dataset.

        If no dataset is assigned, returns ``None``.
        """
        if self.assigned_dataset is None:
            return None
        return self._query.to_dict() != self.assigned_dataset.query.to_dict()

    @staticmethod
    def load_from_serialized_state(state_dict: Dict) -> "Datasource":
        """
        Load a Datasource that was saved with :func:`save_to_file`

        Args:
            state_dict: Serialized JSON object
        """

        state = DatasourceSerializedState.from_dict(state_dict)
        # The json_dataclasses.from_dict() doesn't respect the default value hints, so we fill it out for it
        state.query._fill_out_defaults()

        ds_state = DatasourceState(repo=state.repo, name=state.datasource_name, id=state.datasource_id)
        ds_state.get_from_dagshub()
        ds = Datasource(ds_state)
        ds._query = state.query

        if state.dataset_id is not None:
            ds.load_from_dataset(state.dataset_id, state.dataset_name, change_query=False)

        if state.timestamp is not None:
            ds = ds.as_of(datetime.datetime.fromtimestamp(state.timestamp, tz=datetime.timezone.utc))

        return ds

    def to_voxel51_dataset(self, **kwargs) -> "fo.Dataset":
        """
        Refer to :func:`QueryResult.to_voxel51_dataset() \
        <dagshub.data_engine.model.query_result.QueryResult.to_voxel51_dataset>`\
        for documentation.
        """
        return self.all().to_voxel51_dataset(**kwargs)

    @property
    def default_dataset_location(self) -> Path:
        """
        Default location where datapoint files are stored.

        On UNIX-likes the path is ``~/dagshub/datasets/<repo_name>/<datasource_id>``

        On Windows the path is ``C:\\Users\\<user>\\dagshub\\datasets\\<repo_name>\\<datasource_id>``
        """
        return Path(
            sanitize_filepath(os.path.join(Path.home(), "dagshub", "datasets", self.source.repo, str(self.source.id)))
        )

    def visualize(self, visualizer: Literal["dagshub", "fiftyone"] = "dagshub", **kwargs) -> Union[str, "fo.Session"]:
        """
        Visualize the whole datasource using
        :func:`QueryResult.visualize() <dagshub.data_engine.model.query_result.QueryResult.visualize>`.

        Read the function docs for kwarg documentation.
        """
        if visualizer == "dagshub":
            link = self._generate_visualize_url()

            print("The visualization is available at the following link:")
            print(link)

            webbrowser.open(link)

            return link

        elif visualizer == "fiftyone":
            return self.all().visualize(**kwargs)

    def _generate_visualize_url(self) -> str:
        url: str
        if self.assigned_dataset is not None:
            url = multi_urljoin(
                self.source.repoApi.repo_url, f"datasets/dataset/{self.assigned_dataset.dataset_id}/gallery"
            )
        else:
            url = self.source.url

        full_url = url
        query = self._encode_query_for_frontend()
        if query is not None:
            full_url += f"?{query}"
        return full_url

    def _encode_query_for_frontend(self) -> str:
        """
        Returns a query that is parseable by the frontend.
        It has to be a JSON of the GraphQL query, encoded into b64

        This also omits any null values to make the resulting URL shorter
        """
        params_dict = self._query.to_dict()
        params_dict = {k: v for k, v in params_dict.items() if v is not None}
        params = json.dumps(params_dict)
        params_encoded = base64.urlsafe_b64encode(params.encode("utf-8")).decode("utf-8")
        return f"query={params_encoded}"

    @property
    def fields(self) -> List[MetadataFieldSchema]:
        return self.source.metadata_fields

    @retry(retry=retry_if_exception_type(LSInitializingError), wait=wait_fixed(3), stop=stop_after_attempt(5))
    async def add_annotation_model_from_config(self, config, project_name, ngrok_authtoken, port=9090):
        """
        Initialize a LS backend for ML annotation using a preset configuration.

        Args:
            config: dictionary containing information about the mlflow model, hooks and LS label config
            recommended to use with `get_config()` from preconfigured_models in the orchestrator repo
            project_name: automatically adds backend to project
            ngrok_authtoken: uses ngrok to forward local connection
            port: (optional, default: 9090) port on which orchestrator is hosted
        """
        ls_api_endpoint = multi_urljoin(self.source.repoApi.host, self.source.repo, "annotations/de/api/projects")

        res = requests.get(
            ls_api_endpoint,
            headers={"Authorization": f"Bearer {dagshub.auth.get_token()}"},
        )
        if res.text.startswith("<!DOCTYPE html>"):
            raise LSInitializingError()
        elif res.status_code // 100 != 2:
            raise ValueError(f"Adding backend failed! Response: {res.text}")
        projects = {project["title"]: str(project["id"]) for project in res.json()["results"]}

        if project_name not in projects:
            res = requests.post(
                ls_api_endpoint,
                headers={"Authorization": f"Bearer {dagshub.auth.get_token()}", "Content-Type": "application/json"},
                data=json.dumps({"title": project_name, "label_config": config.pop("label_config")}),
            )
            if res.status_code // 100 != 2:
                raise ValueError(f"Adding backend failed! Response: {res.text}")
        else:
            res = requests.patch(
                multi_urljoin(ls_api_endpoint, projects[project_name]),
                headers={"Authorization": f"Bearer {dagshub.auth.get_token()}", "Content-Type": "application/json"},
                data=json.dumps({"label_config": config.pop("label_config")}),
            )
            if res.status_code // 100 != 2:
                raise ValueError(f"Adding backend failed! Response: {res.text}")

        await self.add_annotation_model(**config, port=port, project_name=project_name, ngrok_authtoken=ngrok_authtoken)

    @retry(retry=retry_if_exception_type(LSInitializingError), wait=wait_fixed(3), stop=stop_after_attempt(5))
    async def add_annotation_model(
        self,
        repo: str,
        name: str,
        version: str = "latest",
        post_hook: Callable[[Any], Any] = lambda x: x,
        pre_hook: Callable[[Any], Any] = lambda x: x,
        port: int = 9090,
        project_name: Optional[str] = None,
        ngrok_authtoken: Optional[str] = None,
    ) -> None:
        """
        Initialize a LS backend for ML annotation.

        Args:
            repo: repository to extract the model from
            name: name of the model in the mlflow registry
            version: (optional, default: 'latest') version of the model in the mlflow registry
            pre_hook: (optional, default: identity function) function that runs before datapoint is sent to the model
            post_hook: (optional, default: identity function) function that converts mlflow model output
            to the desired format
            port: (optional, default: 9090) port on which orchestrator is hosted
            project_name: (optional, default: None) automatically adds backend to project
            ngrok_authtoken: (optional, default: None) uses ngrok to forward local connection
        """

        def fn_encoder(fn):
            return base64.b64encode(cloudpickle.dumps(fn)).decode("utf-8")

        if not ngrok_authtoken and project_name:
            raise ValueError("As `ngrok_authtoken` is not specified, project will have to be added manually.")
        with get_rich_progress() as progress:
            task = progress.add_task("Initializing LS Model...", total=1)
            res = requests.post(
                f"{LS_ORCHESTRATOR_URL}:{port}/configure",
                headers={"Content-Type": "application/json"},
                json=json.dumps(
                    {
                        "host": self.source.repoApi.host,
                        "username": self.source.repoApi.owner,
                        "repo": repo,
                        "model": name,
                        "version": version,
                        "authtoken": dagshub.auth.get_token(),
                        "datasource_repo": self.source.repo,
                        "datasource_name": self.source.name,
                        "pre_hook": fn_encoder(pre_hook),
                        "post_hook": fn_encoder(post_hook),
                    }
                ),
            )
            progress.update(task, advance=1, description="Configured LS model backend container")
            if res.status_code // 100 != 2:
                raise ValueError(f"Adding backend failed! Response: {res.text}")

            if ngrok_authtoken:
                if not self.ngrok_listener:
                    self.ngrok_listener = await ngrok.forward(port, authtoken=ngrok_authtoken)
                endpoint = self.ngrok_listener.url()
            else:
                endpoint = f"{LS_ORCHESTRATOR_URL}:{port}/"
            progress.update(task, advance=1, description="Configured any necessary forwarding")

            if project_name:
                ls_api_endpoint = multi_urljoin(self.source.repoApi.host, self.source.repo, "annotations/de/api")
                res = requests.get(
                    multi_urljoin(ls_api_endpoint, "projects"),
                    headers={"Authorization": f"Bearer {dagshub.auth.get_token()}"},
                )
                if res.text.startswith("<!DOCTYPE html>"):
                    raise LSInitializingError()
                elif res.status_code // 100 != 2:
                    raise ValueError(f"Adding backend failed! Response: {res.text}")
                projects = {project["title"]: str(project["id"]) for project in res.json()["results"]}

                if project_name not in projects:
                    raise ValueError(
                        f"{project_name} not in projects. Available project names: {list(projects.keys())}"
                    )

                res = requests.post(
                    multi_urljoin(ls_api_endpoint, "ml"),
                    headers={"Authorization": f"Bearer {dagshub.auth.get_token()}", "Content-Type": "application/json"},
                    data=json.dumps({"url": endpoint, "project": projects[project_name]}),
                )
                if res.status_code // 100 == 2:
                    progress.update(task, advance=1, description="Added model to LS backend")
                    print("Backend added successfully!")
                else:
                    raise ValueError(f"Adding backend failed! Response: {res.text}")
            else:
                progress.update(task, advance=1, description="Added model to LS backend")
                print(f"Connection Established! Add LS endpoint: {endpoint} to your project.")

    def annotate(self, fields_to_embed=None, fields_to_exclude=None) -> Optional[str]:
        """
        Sends all datapoints in the datasource for annotation in Label Studio.

        Args:
            fields_to_embed: list of meta-data columns that will show up in Label Studio UI.
             if not specified all will be displayed.
            fields_to_exclude: list of meta-data columns that will not show up in Label Studio UI

        .. note::
            This will send ALL datapoints in the datasource for annotation.
            It's recommended to not send a huge amount of datapoints to be annotated at once, to avoid overloading
            the Label Studio workspace.
            Use :func:`QueryResult.annotate() <dagshub.data_engine.model.query_result.QueryResult.annotate>`
            to annotate a result of a query with less datapoints.
            Alternatively, use a lower level :func:`send_datapoints_to_annotation` function

        :return: Link to open Label Studio in the browser
        """
        return self.all().annotate(fields_to_exclude=fields_to_exclude, fields_to_embed=fields_to_embed)

    def send_to_annotation(self):
        """
        deprecated, see :func:`annotate()`

        :meta private:
        """
        return self.annotate()

    def send_datapoints_to_annotation(
        self,
        datapoints: Union[List[Datapoint], "QueryResult", List[Dict]],
        open_project=True,
        ignore_warning=False,
        fields_to_exclude=None,
        fields_to_embed=None,
    ) -> Optional[str]:
        """
        Sends datapoints for annotation in Label Studio.

        Args:
            datapoints: Either of:

                - A :class:`.QueryResult`
                - List of :class:`.Datapoint` objects
                - List of dictionaries. Each dictionary should have fields ``id`` and ``download_url``.
                    ``id`` is the ID of the datapoint in the datasource.

            open_project: Automatically open the created Label Studio project in the browser.
            ignore_warning: Suppress the prompt-warning if you try to annotate too many datapoints at once.
            fields_to_embed: list of meta-data columns that will show up in Label Studio UI.
             if not specified all will be displayed.
            fields_to_exclude: list of meta-data columns that will not show up in Label Studio UI
        Returns:
            Link to open Label Studio in the browser
        """
        for f in (fields_to_embed or []) + (fields_to_exclude or []):
            if not self.has_field(f):
                raise FieldNotFoundError(f)

        if len(datapoints) == 0:
            logger.warning("No datapoints provided to be sent to annotation")
            return None
        elif len(datapoints) > dagshub.common.config.recommended_annotate_limit and not ignore_warning:
            force = prompt_user(
                f"You are attempting to annotate {len(datapoints)} datapoints at once - it's "
                f"recommended to only annotate up to "
                f"{dagshub.common.config.recommended_annotate_limit} "
                f"datapoints at a time."
            )
            if not force:
                return ""

        req_data = {
            "datasource_id": self.source.id,
            "datapoints": [],
            "ls_meta_excludes": fields_to_exclude,
            "ls_meta_includes": fields_to_embed,
        }

        for dp in datapoints:
            req_dict = {}
            if type(dp) is dict:
                req_dict["id"] = dp["datapoint_id"]
                req_dict["download_url"] = dp["download_url"]
            else:
                req_dict["id"] = dp.datapoint_id
                req_dict["download_url"] = dp.download_url
            req_data["datapoints"].append(req_dict)

        init_url = multi_urljoin(self.source.repoApi.data_engine_url, "annotations/init")
        resp = http_request("POST", init_url, json=req_data, auth=self.source.repoApi.auth)

        if resp.status_code != 200:
            logger.error(f"Error while sending request for annotation: {resp.content}")
            return None
        link = resp.json()["link"]

        # Do a raw print so it works in colab/jupyter
        print("Open the following link to start working on your annotation project:")
        print(link)

        if open_project:
            webbrowser.open_new_tab(link)
        return link

    def _launch_annotation_workspace(self):
        try:
            start_workspace_url = multi_urljoin(self.source.repoApi.annotations_url, "start")
            http_request("POST", start_workspace_url, auth=self.source.repoApi.auth)
        except:  # noqa
            pass

    def wait_until_ready(self, max_wait_time=300, fail_on_timeout=True):
        """
        Blocks until the datasource preprocessing is complete.

        Useful when you have just created a datasource and the initial scanning hasn't finished yet.

        Args:
            max_wait_time: Maximum time to wait in seconds
            fail_on_timeout: Whether to raise a RuntimeError or log a warning if the scan does not complete on time
        """

        # Start LS workspace to save time later in the flow
        self._launch_annotation_workspace()

        start = time.time()
        if max_wait_time:
            rich_console.log(f"Maximum waiting time set to {int(max_wait_time / 60)} minutes")
        spinner = rich_console.status("Waiting for datasource preprocessing to complete...")
        with spinner:
            while True:
                self.source.get_from_dagshub()
                if self.source.preprocessing_status == PreprocessingStatus.READY:
                    return

                if self.source.preprocessing_status == PreprocessingStatus.FAILED:
                    raise RuntimeError("Datasource preprocessing failed")

                if max_wait_time is not None and (time.time() - start) > max_wait_time:
                    if fail_on_timeout:
                        raise RuntimeError(
                            f"Time limit of {max_wait_time} seconds reached before processing was completed."
                        )
                    else:
                        logger.warning(
                            f"Time limit of {max_wait_time} seconds reached before processing was completed."
                        )
                        return

                time.sleep(1)

    def has_field(self, field_name: str) -> bool:
        """
        Checks if a metadata field ``field_name`` exists in the datasource.
        """
        reserved_searchable_fields = ["path"]
        fields = (f.name for f in self.fields)
        return field_name in reserved_searchable_fields or field_name in fields

    def __repr__(self):
        res = f"Datasource {self.source.name}"
        res += f"\n\tRepo: {self.source.repo}, path: {self.source.path}"
        if self.assigned_dataset:
            res += (
                f"\n\tAssigned Dataset: {self.assigned_dataset.dataset_name} "
                f"(Query was modified: {self.is_query_different_from_dataset})"
            )
        res += f"\n\t{self._query}"
        res += "\n\tFields:"
        for f in self.fields:
            res += f"\n\t\t{f}"
        return res + "\n"

    """ FUNCTIONS RELATED TO QUERYING
    These are functions that overload operators on the DataSet, so you can do pandas-like filtering
        ds = Dataset(...)
        queried_ds = ds[ds["value"] == 5]
    """

    def __getitem__(self, other: Union[slice, str, "Datasource", "Field"]):
        # Slicing - get items from the slice
        if type(other) is slice:
            return self.sample(other.start, other.stop)

        # Otherwise we're doing querying
        new_ds = self.__deepcopy__()
        if isinstance(other, (str, Field)):
            query_field: Field = Field(other) if isinstance(other, str) else other
            other_query = QueryFilterTree(
                query_field.field_name,
                query_field.as_of_timestamp,
            )
            if not self.has_field(query_field.field_name):
                raise FieldNotFoundError(query_field.field_name)
            if self._query.filter.is_empty:
                new_ds._query.filter = other_query
            else:
                new_ds._query.compose("and", other_query)
            return new_ds
        # "index" is a datasource with a query - return the datasource inside
        # Example:
        #   ds = Dataset()
        #   filtered_ds = ds[ds["aaa"] > 5]
        #   filtered_ds2 = filtered_ds[filtered_ds["bbb"] < 4]
        # filtered_ds2 will be "aaa" > 5 AND "bbb" < 4
        if isinstance(other, Datasource):
            return other

    def __gt__(self, other: object):
        self._test_not_comparing_other_ds(other)
        if not isinstance(other, (int, float, str, datetime.datetime)):
            raise NotImplementedError
        return self.add_query_op("gt", other)

    def __ge__(self, other: object):
        self._test_not_comparing_other_ds(other)
        if not isinstance(other, (int, float, str, datetime.datetime)):
            raise NotImplementedError
        return self.add_query_op("ge", other)

    def __le__(self, other: object):
        self._test_not_comparing_other_ds(other)
        if not isinstance(other, (int, float, str, datetime.datetime)):
            raise NotImplementedError
        return self.add_query_op("le", other)

    def __lt__(self, other: object):
        self._test_not_comparing_other_ds(other)
        if not isinstance(other, (int, float, str, datetime.datetime)):
            raise NotImplementedError
        return self.add_query_op("lt", other)

    def __eq__(self, other: object):
        self._test_not_comparing_other_ds(other)
        if other is None:
            return self.is_null()
        if not isinstance(other, (int, float, str, datetime.datetime)):
            raise NotImplementedError
        return self.add_query_op("eq", other)

    def __ne__(self, other: object):
        self._test_not_comparing_other_ds(other)
        if other is None:
            return self.is_not_null()
        if not isinstance(other, (int, float, str, datetime.datetime)):
            raise NotImplementedError
        return self.add_query_op("eq", other).add_query_op("not")

    def __invert__(self):
        return self.add_query_op("not")

    def __contains__(self, item):
        raise WrongOperatorError("Use `ds.contains(a)` for querying instead of `a in ds`")

    def contains(self, item: str):
        """
        Check if the filtering field contains the specified string item.

        :meta private:
        """
        if type(item) is not str:
            return WrongOperatorError(f"Cannot use contains with non-string value {item}")
        self._test_not_comparing_other_ds(item)
        return self.add_query_op("contains", item)

    def _periodic_filter(self, periodtype, items):
        periods = [str(s) for s in items]
        return self.add_query_op(periodtype, periods)

    def date_field_in_years(self, *item: int):
        """
        Checks if a metadata field (which is of datetime type) is in one of given years list.

        Args:
            List of years.

        Examples::

            datasource[(datasource["y"].date_field_in_years(1979, 2003)

        """

        return self._periodic_filter("year", item)

    def date_field_in_months(self, *item: int):
        """
        Checks if a metadata field (which is of datetime type) is in one of given months list.

        Args:
            List of months.

        Examples::

            datasource[(datasource["y"].date_field_in_months(12, 2)

        """
        return self._periodic_filter("month", item)

    def date_field_in_days(self, *item: int):
        """
        Checks if a metadata field (which is of datetime type) is in one of given days list.

        Args:
            List of days.

        Examples::

            datasource[(datasource["y"].date_field_in_days(25, 2)

        """
        return self._periodic_filter("day", item)

    def date_field_in_timeofday(self, item: str):
        """
        Checks if a metadata field (which is of datetime type) is in given minute range inside the day (any day).
        range is in the format of: "HH:mm-HH:mm" (or "HH:mm:ss-HH:mm:ss") where start hour is on the left.
        a range that starts at one day and ends at next day,
        should be expressed as OR of 2 range filter.

        Args:
            Time range string.

        Examples::

            datasource[(datasource["y"].date_field_in_timeofday("11:30-12:30")

        """
        self._test_not_comparing_other_ds(item)
        return self.add_query_op("timeofday", item)

    def startswith(self, item: str):
        """
        Check if the filtering field starts with the specified string item.

        :meta private:
        """
        if type(item) is not str:
            return WrongOperatorError(f"Cannot use startswith with non-string value {item}")
        return self.add_query_op("startswith", item)

    def endswith(self, item: str):
        """
        Check if the filtering field ends with the specified string item.

        :meta private:
        """
        if type(item) is not str:
            return WrongOperatorError(f"Cannot use endswith with non-string value {item}")
        return self.add_query_op("endswith", item)

    def is_null(self):
        """
        Check if the filtering field is null.

        :meta private:
        """
        field = self._get_filtering_field()
        value_type = metadataTypeLookupReverse[field.valueType.value]
        return self.add_query_op("isnull", value_type())

    def is_not_null(self):
        """
        Check if the filtering field is not null.

        :meta private:
        """
        field = self._get_filtering_field()
        value_type = metadataTypeLookupReverse[field.valueType.value]
        return self.add_query_op("!isnull", value_type())

    def _get_filtering_field(self) -> MetadataFieldSchema:
        field_name = self.get_query().filter.column_filter
        if field_name is None:
            raise RuntimeError("The current query filter is not a field")
        for col in self.source.metadata_fields:
            if col.name == field_name:
                return col
        raise RuntimeError(f"Field {field_name} doesn't exist in the current uploaded metadata")

    def __and__(self, other: "Datasource"):
        return self.add_query_op("and", other)

    def __or__(self, other: "Datasource"):
        return self.add_query_op("or", other)

    # Prevent users from messing up their queries due to operator order
    # They always need to put the dataset query filters in parentheses, otherwise the binary and/or get executed before
    def __rand__(self, other):
        if type(other) is not Datasource:
            raise WrongOrderError(type(other))
        raise NotImplementedError

    def __ror__(self, other):
        if type(other) is not Datasource:
            raise WrongOrderError(type(other))
        raise NotImplementedError

    def add_query_op(
        self,
        op: str,
        other: Optional[Union[str, int, float, "Datasource", "QueryFilterTree", List[str], datetime.datetime]] = None,
    ) -> "Datasource":
        """
        Add a query operation to the current Datasource instance.

        Args:
            op (str): The operation to be performed in the query.
            other (Optional[Union[str, int, float, "Datasource", "DatasourceQuery"]], optional):
                The operand for the query operation. Defaults to None.

        Returns:
            Datasource: A new Datasource instance with the added query operation.

        :meta private:
        """
        new_ds = self.__deepcopy__()
        if type(other) is Datasource:
            other = other.get_query()
        new_ds._query.compose(op, other)
        return new_ds

    @staticmethod
    def _test_not_comparing_other_ds(other):
        if type(other) is Datasource:
            raise DatasetFieldComparisonError()

    @staticmethod
    def _convert_file_to_df(file_path: str):
        # prepare dataframe for import_metadata
        if file_path.lower().endswith(".csv"):
            df = pandas.read_csv(file_path)
        elif file_path.lower().endswith(".parquet"):
            df = pandas.read_parquet(file_path)
        elif file_path.lower().endswith(".zip"):
            df = pandas.read_csv(file_path, compression="zip")
        elif file_path.lower().endswith(".gz"):
            df = pandas.read_csv(file_path, compression="gzip")
        else:
            raise RuntimeError(
                f"File '{file_path}' needs to be a .csv/.parquet or a compressed .zip/.gz to be imported"
            )
        return df

    def import_annotations_from_files(
        self,
        annotation_type: AnnotationType,
        path: Union[str, Path],
        field: str = "imported_annotation",
        load_from: Optional[AnnotationLocation] = None,
        remapping_function: Optional[Callable[[str], str]] = None,
        **kwargs,
    ):
        """
        Imports annotations into the datasource from files

        The annotations will be downloaded and converted into Label Studio tasks,
        that are then uploaded into the specified fields.

        If the annotations are stored in a repo and not locally, they are downloaded to a temporary directory.

        Caveats:
            - YOLO:
                - Images need to also be downloaded to get their dimensions.
                - The .YAML file needs to have the ``path`` argument set to the relative path to the data. \
                    We're using that to download the files
                - You have to specify the ``yolo_type`` kwarg with the type of annotation to import

        Args:
            annotation_type: Type of annotations to import. Possible values are ``yolo`` and ``cvat``
            path: If YOLO - path to the .yaml file, if CVAT - path to the .zip file. \
                Can be either on disk or in repository
            field: Which field to upload the annotations into. \
                If it's an existing field, it has to be a blob field, \
                and it will have the annotations flag set afterwards.
            load_from: Force specify where to get the files from. \
                By default, we're trying to load files from the disk first, and then repository.
                If this is specified, then that check is being skipped and \
                we'll try to download from the specified location.
            remapping_function: Function that maps from a path of the annotation to the path of the datapoint. \
                If None, we try to make a best guess based on the first imported annotation. \
                This might fail, if there is no matching datapoint in the datasource for some annotations \
                or if the paths are wildly different.

        Keyword Args:
            yolo_type: Type of YOLO annotations to import. Either ``bbox``, ``segmentation`` or ``pose``.

        Example to import segmentation annotations into an ``imported_annotations`` field,
        using YOLO information from an ``annotations.yaml`` file (can be local, or in the repo)::

            ds.import_annotations_from_files(
                annotation_type="yolo",
                path="annotations.yaml",
                field="imported_annotations",
                yolo_type="segmentation"
            )
        """

        # Make sure the annotation field exists, is a blob field + has the annotation tag
        existing_fields = [f for f in self.fields if f.name == field]
        if len(existing_fields) != 0:
            f = existing_fields[0]
            if f.valueType != MetadataFieldType.BLOB:
                raise RuntimeError(
                    f"Field {f.name} is not a blob field. "
                    f"Choose a new field or an existing blob field to upload annotations to."
                )
        self.metadata_field(field).set_type(bytes).set_annotation().apply()

        # Run import
        importer = AnnotationImporter(
            ds=self,
            annotations_type=annotation_type,
            annotations_file=path,
            load_from=load_from,
            **kwargs,
        )
        annotation_dict = importer.import_annotations()

        annotation_dict = importer.remap_annotations(annotation_dict, remap_func=remapping_function)

        with rich_console.status("Converting annotations to tasks..."):
            tasks = importer.convert_to_ls_tasks(annotation_dict)

        with self.metadata_context() as ctx:
            for dp, task in tasks.items():
                ctx.update_metadata(dp, {field: task})

        log_message(f'Done! Uploaded annotations for {len(tasks)} datapoints to field "{field}"')


class MetadataContextManager:
    """
    Context manager for updating the metadata on a datasource.
    Batches the metadata changes, so they are being sent all at once.
    """

    def __init__(self, datasource: Datasource):
        self._datasource = datasource
        self._metadata_entries: List[DatapointMetadataUpdateEntry] = []
        self._multivalue_fields = datasource._get_multivalue_fields()

    def update_metadata(self, datapoints: Union[List[str], str], metadata: Dict[str, Any]):
        """
        Update metadata for the specified datapoints.

        .. note::
            If ``datapoints`` is a list, the same metadata is assigned to all the datapoints in the list.
            Call ``update_metadata()`` separately for each datapoint if you need to assign different metadata.

        Args:
            datapoints (Union[List[str], str]): A list of datapoints or a single datapoint path to update metadata for.
            metadata (Dict[str, Any]): A dictionary containing metadata key-value pairs to update.

        Example::

            with ds.metadata_context() as ctx:
                metadata = {
                    "episode": 5,
                    "has_baby_yoda": True,
                }

                # Attach metadata to a single specific file in the datasource.
                # The first argument is the filepath to attach metadata to, **relative to the root of the datasource**.
                ctx.update_metadata("images/005.jpg", metadata)

                # Attach metadata to several files at once:
                ctx.update_metadata(["images/006.jpg","images/007.jpg"], metadata)

        """
        if isinstance(datapoints, str):
            datapoints = [datapoints]

        field_value_types = {f.name: f.valueType for f in self._datasource.fields}
        document_fields = set(self._datasource.document_fields)

        for dp in datapoints:
            for k, v in metadata.items():
                if v is None:
                    continue
                if k in autogenerated_columns:
                    continue

                if isinstance(v, list):
                    if k not in self._multivalue_fields:
                        self._multivalue_fields.add(k)
                        # Promote all existing ones to multivalue
                        for e in self._metadata_entries:
                            if e.key == k:
                                e.allowMultiple = True
                    for sub_val in v:
                        time_zone = None
                        value_type = field_value_types.get(k)
                        if value_type is None:
                            value_type = metadataTypeLookup[type(sub_val)]
                            field_value_types[k] = value_type
                        # Don't override bytes if they're not bytes and not documents
                        # - probably just undownloaded values
                        if value_type == MetadataFieldType.BLOB and not isinstance(sub_val, bytes):
                            if k not in document_fields:
                                continue
                        if isinstance(v, str) and k in document_fields:
                            v = v.encode("utf-8")
                        if isinstance(v, datetime.datetime):
                            time_zone = _get_datetime_utc_offset(v)
                            v = int(v.timestamp() * 1000)
                        if isinstance(v, bytes):
                            sub_val = wrap_bytes(sub_val)
                        self._metadata_entries.append(
                            DatapointMetadataUpdateEntry(
                                url=dp,
                                key=k,
                                value=str(sub_val),
                                # todo: preliminary type check
                                valueType=value_type,
                                allowMultiple=k in self._multivalue_fields,
                                timeZone=time_zone,
                            )
                        )

                else:
                    value_type = field_value_types.get(k)
                    if value_type is None:
                        value_type = metadataTypeLookup[type(v)]
                        field_value_types[k] = value_type
                    # Don't override bytes if they're not bytes - probably just undownloaded values
                    if value_type == MetadataFieldType.BLOB and not isinstance(v, bytes):
                        if k not in document_fields:
                            continue

                    time_zone = None
                    if isinstance(v, str) and k in document_fields:
                        v = v.encode("utf-8")
                    if isinstance(v, bytes):
                        v = wrap_bytes(v)
                    if isinstance(v, datetime.datetime):
                        time_zone = _get_datetime_utc_offset(v)
                        v = int(v.timestamp() * 1000)

                    self._metadata_entries.append(
                        DatapointMetadataUpdateEntry(
                            url=dp,
                            key=k,
                            value=str(v),
                            valueType=value_type,
                            # todo: preliminary type check
                            allowMultiple=k in self._multivalue_fields,
                            timeZone=time_zone,
                        )
                    )

    def get_metadata_entries(self):
        return self._metadata_entries

    def clear(self):
        self._metadata_entries.clear()

    def __len__(self):
        return len(self._metadata_entries)


def _get_datetime_utc_offset(t):
    """
    return a timezone offset in the form of "+03:00" or "-03:00"
    """

    if t.tzinfo is None:
        return None

    offset = t.utcoffset()
    if offset is None:
        return None

    # Format the offset as a string
    offset_hours = int(offset.total_seconds() // 3600)
    offset_minutes = int((offset.total_seconds() % 3600) // 60)
    offset_str = f"{offset_hours:+03d}:{offset_minutes:02d}"
    return offset_str


@dataclass
class DatasourceQuery(DataClassJsonMixin):
    as_of: Optional[int] = field(default=None, metadata=config(exclude=exclude_if_none, letter_case=LetterCase.CAMEL))
    time_zone: Optional[str] = field(
        default=None, metadata=config(exclude=exclude_if_none, letter_case=LetterCase.CAMEL)
    )
    select: Optional[List[Dict]] = field(default=None, metadata=config(exclude=exclude_if_none))
    filter: "QueryFilterTree" = field(
        default=QueryFilterTree(),
        metadata=config(field_name="query", encoder=QueryFilterTree.serialize, decoder=QueryFilterTree.deserialize),
    )
    order_by: Optional[List] = field(
        default=None, metadata=config(exclude=exclude_if_none, letter_case=LetterCase.CAMEL)
    )

    def _fill_out_defaults(self):
        """For functions that don't utilize the default hints of the dataclass"""
        self.filter = QueryFilterTree()

    def __deepcopy__(self, memodict={}):
        other = DatasourceQuery(
            as_of=self.as_of,
            time_zone=self.time_zone,
            filter=self.filter.__deepcopy__(),
        )
        if self.select is not None:
            other.select = self.select.copy()
        if self.order_by is not None:
            other.order_by = self.order_by.copy()
        return other

    def compose(
        self, op: str, other: Optional[Union[str, int, float, "DatasourceQuery", "QueryFilterTree", "Datasource"]]
    ):
        other_query: Optional["DatasourceQuery"] = None
        # Extract the filter tree for composition
        if isinstance(other, (DatasourceQuery, Datasource)):
            if type(other) is Datasource:
                other = other.get_query()
            other_query = other
            other = other.filter

        self.filter.compose(op, other)

        # If composition went successfully, carry over the as_of, select and order_by from the other query
        if other_query is not None:
            if other_query.select is not None:
                self.select = other_query.select
            if other_query.as_of is not None:
                self.as_of = other_query.as_of
            if other_query.time_zone is not None:
                self.time_zone = other_query.time_zone
            if other_query.order_by is not None:
                self.order_by = other_query.order_by


@dataclass
class DatasourceSerializedState(DataClassJsonMixin):
    """
    Serialized state of the datasource.
    This should be enough to recreate the exact copy of the datasource back (with additional requests)

    Also carries additional information that might be useful for the user:
        - if the state of the datasource at the point of saving differed from the dataset
        - the timestamp of saving
        - link to open the datasource on DagsHub
    """

    repo: str
    """Repository this datasource is on"""
    datasource_id: Union[str, int]
    """ID of the datasource"""
    datasource_name: str
    """Name of the datasource"""
    query: DatasourceQuery
    """Query at the time of saving"""
    dataset_id: Optional[Union[str, int]] = None
    """ID of the assigned dataset"""
    dataset_name: Optional[str] = None
    """Name of the assigned dataset"""
    timestamp: Optional[float] = None
    """Timestamp of serialization"""
    modified: Optional[bool] = None
    """Does the query differ from the query in the assigned dataset"""
    link: Optional[str] = None
    """URL to open this datasource on DagsHub"""
    timed_link: Optional[str] = None
    """URL to open this datasource with the data at the time of querying"""


@dataclass
class DatasetState:
    """
    Information about the Dataset.

    Dataset is a Datasource with a Query applied on it.
    """

    dataset_id: Union[str, int]
    """
    ID of the dataset
    """
    dataset_name: str
    """
    Name of the dataset
    """
    datasource_id: Union[str, int]
    """
    ID of the datasource with which this dataset is associated
    """
    query: DatasourceQuery
    """
    Query of this dataset
    """

    @staticmethod
    def from_dataset_query(
        dataset_id: Union[str, int], dataset_name: str, datasource_id: Union[str, int], dataset_query: Union[Dict, str]
    ) -> "DatasetState":
        if type(dataset_query) is str:
            dataset_query = json.loads(dataset_query)
        query = DatasourceQuery.from_dict(dataset_query)
        res = DatasetState(dataset_id=dataset_id, dataset_name=dataset_name, datasource_id=datasource_id, query=query)
        return res

    @staticmethod
    def from_gql_dataset_result(dataset_result: "DatasetResult") -> "DatasetState":
        return DatasetState.from_dataset_query(
            dataset_id=dataset_result.id,
            dataset_name=dataset_result.name,
            datasource_id=dataset_result.datasource.id,
            dataset_query=dataset_result.datasetQuery,
        )



from abc import abstractmethod
from pathlib import Path, PurePosixPath
from typing import Literal, Optional

from dagshub.common.api import RepoAPI
from dagshub.common.helpers import log_message
from dagshub.streaming import DagsHubFilesystem


class ModelLoader:
    def __init__(self, repo_api: RepoAPI):
        self.repo_api = repo_api

    def load_model(
        self,
        mode: Literal["eager", "lazy"],
        download_dest: Path,
        revision: Optional[str] = None,
    ) -> Path:
        if mode == "eager":
            return self._eager_load(download_dest)
        elif mode == "lazy":
            if revision is None:
                revision = self.repo_api.default_branch
            return self._lazy_load(download_dest, revision)
        raise RuntimeError(f"Unknown model load mode [{mode}]")

    @abstractmethod
    def _eager_load(self, download_dest: Path) -> Path: ...

    def _lazy_load(self, download_dest: Path, revision: str) -> Path:
        fs = DagsHubFilesystem(
            project_root=download_dest,
            repo_url=self.repo_api.repo_url,
            branch=revision,
            frameworks=["transformers"],
        )
        res_path = download_dest / self.model_path
        log_message(
            f"Running install_hooks() in dir {download_dest}, "
            f"run dagshub.streaming.uninstall_hooks() if you want to run another model later"
        )
        fs.install_hooks()
        return res_path

    @property
    @abstractmethod
    def model_path(self) -> Path: ...


class RepoModelLoader(ModelLoader):
    def __init__(self, repo_api: RepoAPI, revision: str, path: PurePosixPath):
        super().__init__(repo_api)
        self.revision = revision
        self.path = path

    def load_model(
        self,
        mode: Literal["eager", "lazy"],
        download_dest: Path,
        revision: Optional[str] = None,
    ) -> Path:
        """Overrides the original to change the revision used. If it's not set, use the revision from the class"""
        if revision is None:
            revision = self.revision
        return super().load_model(mode, download_dest, revision)

    def _eager_load(self, download_dest: Path) -> Path:
        remote_path = self.path
        local_path = download_dest / self.model_path
        self.repo_api.download(
            remote_path, local_path=local_path, revision=self.revision
        )
        return local_path

    @property
    def model_path(self) -> Path:
        return Path(self.path)


class BucketModelLoader(ModelLoader):
    def __init__(self, repo_api: RepoAPI, path: PurePosixPath):
        super().__init__(repo_api)
        self.path = path

    def _eager_load(self, download_dest: Path) -> Path:
        # Need to change from s3/bucket/bla-bla to s3:/bucket/bla-bla for download to work
        remote_path = str(self.path).replace("/", ":/", 1)
        local_path = download_dest / self.model_path
        self.repo_api.download(remote_path, local_path=local_path)
        return local_path

    @property
    def model_path(self) -> Path:
        return Path(".dagshub") / "storage" / self.path


class DagsHubStorageModelLoader(BucketModelLoader):
    def __init__(self, repo_api: RepoAPI, path: PurePosixPath):
        bucket_path = PurePosixPath("s3") / repo_api.repo_name / path
        super().__init__(repo_api, bucket_path)


class MLflowArtifactModelLoader(ModelLoader):
    def _eager_load(self, download_dest: Path) -> Path:
        raise NotImplementedError

    @property
    def model_path(self) -> Path:
        raise NotImplementedError



import datetime
import logging
import os
import threading
import traceback
from typing import Optional, Dict, List, Set, Union, cast

import yaml
from httpx import Auth

from dagshub.auth import oauth
from dagshub.auth.token_auth import (
    HTTPBearerAuth,
    DagshubTokenABC,
    TokenDeserializationError,
    AppDagshubToken,
    EnvVarDagshubToken,
    DagshubAuthenticator,
)
from dagshub.common import config
from dagshub.common.helpers import http_request, log_message, prompt_user
from dagshub.common.util import multi_urljoin

logger = logging.getLogger(__name__)

APP_TOKEN_TYPE = "app-token"


class InvalidTokenError(Exception):
    def __str__(self):
        print("The token is not a valid DagsHub token")


_validated_env_tokens = set()


class TokenStorage:
    def __init__(self, cache_location: Optional[str] = None, **kwargs):
        cache_location = cache_location or config.cache_location
        self.cache_location = cache_location
        self.schema_version = config.TOKENS_CACHE_SCHEMA_VERSION
        self.__token_cache: Optional[Dict[str, List[DagshubTokenABC]]] = None

        # We check tokens only once for validity, so we don't do a lot of redundant requests
        #   maybe there is a point to re-evaluate them once in a while
        self._known_good_tokens: Dict[str, Set[DagshubTokenABC]] = {}

        self.__token_access_lock = threading.RLock()
        self._accessing_as_was_printed = False

    @property
    def _token_cache(self):
        if self.__token_cache is None:
            self.__token_cache = self._load_cache_file()
            self.remove_expired_tokens()
        return self.__token_cache

    @property
    def _token_access_lock(self):
        if not hasattr(self, "__token_access_lock"):
            self.__token_access_lock = threading.RLock()
        return self.__token_access_lock

    def remove_expired_tokens(self):
        had_changes = False
        for host, tokens in self._token_cache.items():
            if host == "version":
                continue
            expired_tokens = filter(lambda token: token.is_expired, tokens)
            for t in expired_tokens:
                had_changes = True
                tokens.remove(t)
        if had_changes:
            logger.info("Removed expired tokens from the token cache")
            self._store_cache_file()

    def add_token(self, token: Union[str, DagshubTokenABC], host: Optional[str] = None, skip_validation=False):
        host = host or config.host

        if isinstance(token, str):
            token = AppDagshubToken(token)

        token = cast(DagshubTokenABC, token)

        if self._token_already_exists(token.token_text, host):
            logger.warning("The added token already exists in the token cache, skipping")
            return

        if not skip_validation:
            if not TokenStorage.is_valid_token(token.token_text, host):
                raise InvalidTokenError

        if host not in self._token_cache:
            self._token_cache[host] = []
        self._token_cache[host].append(token)
        self._store_cache_file()

    def invalidate_token(self, token: DagshubTokenABC, host: Optional[str] = None):
        host = host or config.host

        try:
            if host in self._token_cache:
                tokens = self._token_cache[host]
                tokens.remove(token)

                if host in self._known_good_tokens:
                    good_token_set = self._known_good_tokens[host]
                    if token in good_token_set:
                        good_token_set.remove(token)

                self._store_cache_file()
        except ValueError:
            logger.warning(f"Token {token} does not exist in the storage")

    def get_authenticator(
        self, host: Optional[str] = None, fail_if_no_token: bool = False, **kwargs
    ) -> DagshubAuthenticator:
        """
        Returns the authenticator object, that can renegotiate tokens in case of failure
        """
        host = host or config.host
        token = self.get_token_object(host, fail_if_no_token, **kwargs)
        return DagshubAuthenticator(token, token_storage=self, host=host)

    def get_token_object(self, host: Optional[str] = None, fail_if_no_token: bool = False, **kwargs) -> DagshubTokenABC:
        """
        This function does following:
        - Iterates over all tokens in the cache for the provided host
        - Finds a first valid token and returns it
        - If it finds an invalid token, it deletes it from the cache

        We're using a set of known good tokens to skip rechecking for token validity every time
        """

        host = host or config.host
        if host == config.host and config.token is not None:
            if config.token not in _validated_env_tokens:
                user = TokenStorage.get_username_of_token(config.token, host)
                if user is not None:
                    self._print_accessing_as(user)
                    _validated_env_tokens.add(config.token)
                else:
                    raise RuntimeError("Provided DagsHub token is not valid")
            return EnvVarDagshubToken(config.token, host)

        with self._token_access_lock:
            tokens = self._token_cache.get(host, [])

            if host not in self._known_good_tokens:
                self._known_good_tokens[host] = set()
            good_token_set = self._known_good_tokens[host]
            good_token = None
            good_user = None
            token_queue = list(sorted(tokens, key=lambda t: t.priority))

            for token in token_queue:
                if token.is_expired:
                    self.invalidate_token(token, host)
                    good_token_set = self._known_good_tokens[host]

                if token in good_token_set:
                    good_token = token
                    break
                # Check token validity
                user = TokenStorage.get_username_of_token(token, host)
                if user is not None:
                    good_token = token
                    good_token_set.add(token)
                    good_user = user
                # Remove invalid token from the cache
                else:
                    self.invalidate_token(token, host)
                    good_token_set = self._known_good_tokens[host]
                if good_token is not None:
                    break

            # Couldn't manage to find a good token after the search
            # Either go through the oauth flow, or throw a runtime error
            if good_token is None:
                if fail_if_no_token:
                    raise RuntimeError(
                        f"No valid tokens found for host '{host}'.\n"
                        "Log into DagsHub by executing `dagshub login` in your terminal"
                    )
                else:
                    logger.debug(f"No valid tokens found for host '{host}'. Authenticating with OAuth")
                    good_token = oauth.oauth_flow(host, **kwargs)
                    tokens.append(good_token)
                    good_token_set.add(good_token)
                    good_user = TokenStorage.get_username_of_token(good_token, host)
                    # Save the cache
                    self._token_cache[host] = tokens
                    self._store_cache_file()

            self._print_accessing_as(good_user)
            return good_token

    def get_token(self, host: Optional[str] = None, fail_if_no_token: bool = False, **kwargs) -> str:
        """
        Return the raw token string
        This is a lower level method that cannot do renegotiations, we only return the token itself here.
        Used mainly for setting environment variables, for example for MLflow
        """
        return self.get_token_object(host, fail_if_no_token).token_text

    def _token_already_exists(self, token_text: str, host: str):
        for token in self._token_cache.get(host, []):
            if token.token_text == token_text:
                return True
        return False

    @staticmethod
    def _is_expired(token: Dict[str, str]) -> bool:
        if "expiry" not in token:
            return True
        if token["expiry"] == "never":
            return False
        # Need to cut off the three additional precision numbers in milliseconds, because %f only parses 6 digits
        expiry = token["expiry"][:-4] + "Z"
        expiry_dt = datetime.datetime.strptime(expiry, "%Y-%m-%dT%H:%M:%S.%fZ")
        is_expired = expiry_dt < datetime.datetime.utcnow()
        return is_expired

    @staticmethod
    def get_username_of_token(token: Union[str, Auth, DagshubTokenABC], host: str) -> Optional[Dict]:
        """
        Check for token validity and return the dictionary with the info of the user of the token

        Args:
            token: token to check validity
            host: which host to connect against
        """
        host = host or config.host
        check_url = multi_urljoin(host, "api/v1/user")
        if type(token) is str:
            auth = HTTPBearerAuth(token)
        else:
            auth = token
        resp = http_request("GET", check_url, auth=auth)

        try:
            assert resp.status_code == 200
            user = resp.json()
            assert "login" in user
            assert "username" in user
            return user
        except AssertionError:
            return None

    @staticmethod
    def is_valid_token(token: Union[str, Auth, DagshubTokenABC], host: str) -> bool:
        """
        Check for token validity

        Args:
            token: token to check validity
            host: which host to connect against
        """
        return TokenStorage.get_username_of_token(token, host) is not None

    def _load_cache_file(self) -> Dict[str, List[DagshubTokenABC]]:
        logger.debug(f"Loading token cache from {self.cache_location}")
        if not os.path.exists(self.cache_location):
            logger.debug("Token cache file doesn't exist")
            return {}
        try:
            with open(self.cache_location) as f:
                cache_yaml = yaml.load(f, yaml.Loader)
                version = cache_yaml.get("version", "1")
                if version == "1":
                    return self._v1_token_list_parser(cache_yaml)
                raise RuntimeError(f"Don't know how to parse token schema {version}")
        except Exception:
            logger.error(f"Error while loading DagsHub token cache: {traceback.format_exc()}")
            raise

    @staticmethod
    def _v1_token_list_parser(cache_yaml: Dict[str, Union[str, List[Dict]]]) -> Dict[str, List[DagshubTokenABC]]:
        res = {}

        token_class_map = {}
        for token_class in DagshubTokenABC.__subclasses__():
            token_class_map[token_class.token_type] = token_class

        for host, tokens in cache_yaml.items():
            if host == "version":
                continue
            if len(tokens) == 0:
                continue
            host_tokens = []
            for token_dict in tokens:
                try:
                    token = token_class_map[token_dict["token_type"]].deserialize(token_dict)
                    host_tokens.append(token)
                except TokenDeserializationError as e:
                    logger.warning(f"Failed to deserialize token {token_dict}: {e}")
            res[host] = host_tokens
        return res

    def _store_cache_file(self):
        logger.debug(f"Dumping token cache to {self.cache_location}")
        try:
            dirpath = os.path.dirname(self.cache_location)
            if not os.path.exists(dirpath):
                os.makedirs(dirpath)
            dict_to_dump = {"version": self.schema_version}
            for host, tokens in self.__token_cache.items():
                dict_to_dump[host] = [t.serialize() for t in tokens]
            with open(self.cache_location, "w") as f:
                yaml.dump(dict_to_dump, f, yaml.Dumper)
        except Exception:
            logger.error(f"Error while storing DagsHub token cache: {traceback.format_exc()}")
            raise

    def _print_accessing_as(self, user: Dict):
        """
        This function prints a message to the log that we are accessing as a certain user.
        It does this only once per command, to avoid spamming the logs.
        It called after successful token validation.

        """
        if self._accessing_as_was_printed:
            return

        username = user["username"]

        log_message(f"Accessing as {username}")
        self._accessing_as_was_printed = True

    def __getstate__(self):
        d = self.__dict__
        # Don't pickle the lock. This will make it so multiple authenticators might request for tokens at the same time
        # This can lead to e.g. multiple OAuth requests firing at the same time, which is not desirable
        # However, I'm not sure of a good way to solve it
        access_lock_key = f"_{self.__class__.__name__}__token_access_lock"
        if access_lock_key in d:
            del d[access_lock_key]
        return d

    def __setstate__(self, state):
        self.__dict__ = state


_token_storage: Optional[TokenStorage] = None


def _get_token_storage(**kwargs):
    global _token_storage
    if _token_storage is None:
        _token_storage = TokenStorage(**kwargs)
    return _token_storage


def get_authenticator(**kwargs) -> DagshubAuthenticator:
    """
    Get an authenticator object that can be used to authenticate a request to DagsHub using an http library like
    ``httpx`` or ``requests``.

    When used with ``httpx``, the authenticator has renegotiation logic. That means that if DagsHub rejects a token,
    the authenticator will try to get another token from the cache, or fall back to OAuth.

    If no valid token was found and ``fail_if_no_token`` wasn't set to ``True``, triggers OAuth flow.

    Keyword Args:
        fail_if_no_token: Whether to get an OAuth token if no valid token was found in cache initially.
            If set to ``True``, raises a ``RuntimeError``.
            If set to ``False`` (default), launches the OAuth flow.
        cache_location: Path to an alternative cache location.
            You can override the default cache location to be used by the client by setting the
            ``DAGSHUB_CLIENT_TOKENS_CACHE`` environment variable.
        host: URL of the hosted DagsHub instance. default is ``https://dagshub.com``.

    """
    return _get_token_storage(**kwargs).get_authenticator(**kwargs)


def get_token_object(**kwargs):
    """
    Gets a DagsHub token object, by default if no token is found authenticates with OAuth.
    The token object has additional information about the type and expiry of the token.

    Keyword Args:
        fail_if_no_token: Whether to get an OAuth token if no valid token was found in cache initially.
            If set to ``True``, raises a ``RuntimeError``.
            If set to ``False`` (default), launches the OAuth flow.
        cache_location: Path to an alternative cache location.
            You can override the default cache location to be used by the client by setting the
            ``DAGSHUB_CLIENT_TOKENS_CACHE`` environment variable.
        host: URL of the hosted DagsHub instance. default is ``https://dagshub.com``.

    """
    return _get_token_storage(**kwargs).get_token_object(**kwargs)


def get_token(**kwargs) -> str:
    """
    Gets a DagsHub token in regular string form.

    Use this function when you want to, for example,
    manually authenticate with DagsHub's MLflow using ``MLFLOW_TRACKING_PASSWORD``.

    If no valid token was found and ``fail_if_no_token`` wasn't set to ``True``, triggers OAuth flow.


    Keyword Args:
        fail_if_no_token: Whether to get an OAuth token if no valid token was found in cache initially.
            If set to ``True``, raises a ``RuntimeError``.
            If set to ``False`` (default), launches the OAuth flow.
        cache_location: Path to an alternative cache location.
            You can override the default cache location to be used by the client by setting the
            ``DAGSHUB_CLIENT_TOKENS_CACHE`` environment variable.
        host: URL of the hosted DagsHub instance. default is ``https://dagshub.com``.

    """
    return _get_token_storage(**kwargs).get_token(**kwargs)


def add_app_token(token: str, host: Optional[str] = None, **kwargs):
    """
    Adds an application token to the token cache.
    This is a long-lived token that you can add/revoke in your profile settings on DagsHub.

    Args:
        token: Token value
        host: URL of the hosted DagsHub instance. Leave empty to use the default ``https://dagshub.com``.

    Keyword Args:
        cache_location: Path to an alternative cache location.
            You can override the default cache location to be used by the client by setting the
            ``DAGSHUB_CLIENT_TOKENS_CACHE`` environment variable.
    """
    token_obj = AppDagshubToken(token)
    _get_token_storage(**kwargs).add_token(token_obj, host)


def add_oauth_token(host: Optional[str] = None, referrer: Optional[str] = None, **kwargs):
    """
    Launches the OAuth flow that generates a short-lived token.

    .. note::
        This will open a new browser window, so this is not a CI/headless friendly function.
        Consider using :func:`.add_app_token` or setting the ``DAGSHUB_USER_TOKEN`` env var in those cases.

    Args:
        host: URL of the hosted DagsHub instance. Leave empty to use the default ``https://dagshub.com``.
        referrer: For custom referral flows

    Keyword Args:
        cache_location: Path to an alternative cache location.
            You can override the default cache location to be used by the client by setting the
            ``DAGSHUB_CLIENT_TOKENS_CACHE`` environment variable.
    """
    host = host or config.host
    token = oauth.oauth_flow(host, referrer=referrer)
    _get_token_storage(**kwargs).add_token(token, host, skip_validation=True)


def clear_token_cache(force=False):
    """
    Clear the token cache file.

    .. warning::
        This will remove ALL tokens you have stored for any dagshub instance you connected to.
    """
    cache = config.cache_location
    if not force:
        prompt = prompt_user(f"This will remove ALL tokens stored in the cache file at {cache}.")
        if not prompt:
            return

    os.remove(cache)
    print(f"Removed all tokens stored at {cache}")



import json
import logging
from typing import Optional, Union, List, TYPE_CHECKING, Dict

from dagshub.common.analytics import send_analytics_event
from dagshub.common.api.repo import RepoAPI
from dagshub.common.util import lazy_load, removeprefix
from dagshub.data_engine.client.data_client import DataClient
from dagshub.data_engine.model.datasource import Datasource
from dagshub.data_engine.model.datasource_state import DatasourceState, DatasourceType, path_regexes
from dagshub.data_engine.model.errors import DatasourceNotFoundError

if TYPE_CHECKING:
    import mlflow
    import mlflow.artifacts as mlflow_artifacts
    import mlflow.entities
else:
    mlflow = lazy_load("mlflow")
    mlflow_artifacts = lazy_load("mlflow.artifacts", "mlflow")

logger = logging.getLogger(__name__)


def create_datasource(repo: str, name: str, path: str, revision: Optional[str] = None) -> Datasource:
    """
    Create a datasource from a path in the repo or a storage bucket URL.
    You can have multiple datasources pointing at the same path.

    Args:
        repo: Repo in ``<owner>/<reponame>`` format
        name: Name of the datasource to be created. Name should be unique across the repository's datasources
        path: Either of:

            - a path to a directory inside the Git/DVC repo on DagsHub:
                ``path/to/dir``
            - URL pointing to a storage bucket which is connected to the DagsHub repo:
                ``s3://bucketname/path/in/bucket``
        revision: Branch or revision the datasource should be used with.
                    Only valid when using a Git/DVC path inside the DagsHub repo.
                    The default repo branch is used if this is left blank.

    Returns:
        Datasource: Created datasource

    Raises:
        DatasourceAlreadyExistsError: Datasource with this name already exists in repo.

    """

    if path_regexes[DatasourceType.BUCKET].fullmatch(path):
        if revision is not None:
            raise ValueError("revision cannot be used together with bucket URLs")
        return create_from_bucket(repo, name, bucket_url=path)
    else:
        return create_from_repo(repo, name, path=path, revision=revision)


def create(*args, **kwargs) -> Datasource:
    """Alias for :func:`create_datasource`"""
    return create_datasource(*args, **kwargs)


def get_or_create(repo: str, name: str, path: str, revision: Optional[str] = None) -> Datasource:
    """
    First attempts to get the repo datasource with the given name, and only if that fails,
    invokes create_datasource with the given parameters.
    See the docs on create_datasource for more info.



    Args:
        repo (str): Repo in the format of `user/repo`
        name (str): The name of the datasource to retrieve or create.
        path (str): The path to the datasource within the repository.
        revision (Optional[str], optional): The specific revision or version of the datasource to retrieve.

    Returns:
        Datasource: The retrieved or newly created Datasource instance.

    Raises:
        DatasourceAlreadyExistsError: Datasource with this name already exists in repo.
    """
    try:
        return get_datasource(repo, name)
    except DatasourceNotFoundError:
        return create_datasource(repo, name, path, revision)


def create_from_bucket(repo: str, name: str, bucket_url: str) -> Datasource:
    """
    :meta private:
    """
    # TODO: validation
    source = _create_datasource_state(repo, name, DatasourceType.BUCKET, bucket_url)
    return Datasource(source)


def create_from_repo(repo: str, name: str, path: str, revision: Optional[str] = None) -> Datasource:
    """
    :meta private:
    """
    if revision is None:
        repo_api = RepoAPI(repo)
        revision = repo_api.default_branch
    url = f"repo://{repo}/{revision}:{path.lstrip('/')}"
    source = _create_datasource_state(repo, name, DatasourceType.REPOSITORY, url)
    if revision is not None:
        source.revision = revision
    return Datasource(source)


def get_datasource(repo: str, name: Optional[str] = None, id: Optional[Union[int, str]] = None, **kwargs) -> Datasource:
    """
    Gets datasource with matching name or id for the repo

    Args:
        repo: Repo in ``<owner>/<reponame>`` format
        name: Name of the datasource
        id: ID of the datasource

    Kwargs:
        revision - for repo datasources defines which branch/revision to download from.
        If not specified, uses the default branch of the repo

    Returns:
        Datasource: datasource that has supplied name and/or id

    Raises:
        DatasourceNotFoundError: The datasource with this id or name does not exist.

    """
    ds_state = DatasourceState(repo=repo, name=name, id=id)
    ds_state.get_from_dagshub()
    if "revision" in kwargs:
        ds_state.revision = kwargs["revision"]
    return Datasource(ds_state)


def get_datasource_from_file(path: str) -> Datasource:
    """
    [EXPERIMENTAL]
    Load a datasource from a local file

    Args:
        path: Path to the ``.dagshub`` file with the relevant datasource

    Returns:
        ds: Datasource that was logged to the file
    """
    with open(path, "r") as file:
        state = json.load(file)
    return Datasource.load_from_serialized_state(state)


def get_datasources(repo: str) -> List[Datasource]:
    """
    Get all datasources that exist on the repo

    Args:
        repo: Repo in ``<owner>/<reponame>`` format

    Returns:
        list(Datasource): All datasources that exist for the repository
    """
    send_analytics_event("Client_DataEngine_getDatasources")
    client = DataClient(repo)
    sources = client.get_datasources(None, None)
    return [Datasource(DatasourceState.from_gql_result(repo, source)) for source in sources]


def get_from_mlflow(
    run: Optional[Union["mlflow.entities.Run", str]] = None, artifact_name: Optional[str] = None
) -> Dict[str, Datasource]:
    """
    Load datasources from an MLflow run.

    To save a datasource to MLflow, use
    :func:`QueryResult.log_to_mlflow()<dagshub.data_engine.model.query_result.QueryResult.log_to_mlflow>`.

    Args:
        run: MLflow Run or its ID to load the datasource from.
            If ``None``, loads datasource from the current active run.
        artifact_name: Name of the datasource artifact in the run.
            If specified, will only return the datasource defined in this artifact.

    Returns:
        Dictionary where keys are the artifacts, and the values are the datasource stored in the artifact.
    """
    mlflow_run: "mlflow.entities.Run"
    if run is None:
        mlflow_run = mlflow.active_run()
    elif type(run) is str:
        mlflow_run = mlflow.get_run(run)
    else:
        mlflow_run = run

    return _load_datasources_from_run(mlflow_run, artifact_name)


def get(*args, **kwargs) -> Datasource:
    """Alias for :func:`get_datasource`"""
    return get_datasource(*args, **kwargs)


def _create_datasource_state(repo: str, name: str, source_type: DatasourceType, path: str) -> DatasourceState:
    ds = DatasourceState(name=name, repo=repo)
    ds.source_type = source_type
    ds.path = path
    ds.create()
    return ds


def _load_datasources_from_run(
    run: "mlflow.entities.Run", artifact_name: Optional[str] = None
) -> Dict[str, Datasource]:
    if artifact_name is not None:
        artifact_uri: str = run.info.artifact_uri
        artifact_path = f"{artifact_uri.rstrip('/')}/{artifact_name.lstrip('/')}"

        ds_states = {artifact_name: mlflow_artifacts.load_dict(artifact_path)}
    else:
        # Load all artifacts ending with ".dagshub.dataset.json"
        artifacts = mlflow_artifacts.list_artifacts(run.info.artifact_uri)
        # mlflow returns paths with `artifacts/` in the beginning. Need to cut it off because it's also in artifact_uri
        artifact_paths = [removeprefix(a.path, "artifacts/") for a in artifacts]

        ds_states = {
            p: mlflow_artifacts.load_dict(f"{run.info.artifact_uri}/{p}")
            for p in artifact_paths
            if p.endswith(".dagshub.dataset.json")
        }

    return {p: Datasource.load_from_serialized_state(ds_state) for p, ds_state in ds_states.items()}


__all__ = [
    create_datasource.__name__,
    create.__name__,
    create_from_bucket.__name__,
    create_from_repo.__name__,
    get_datasource.__name__,
    get_datasources.__name__,
    get.__name__,
    get_or_create.__name__,
    get_from_mlflow.__name__,
]



from functools import cached_property
from typing import Optional, Any, Union, Dict

import dacite

from dagshub.auth import get_authenticator, get_token
from dagshub.auth.token_auth import HTTPBearerAuth
from dagshub.common import config
from dagshub.common.api.responses import UserAPIResponse
from dagshub.common.helpers import http_request
from dagshub.common.util import multi_urljoin


class UserNotFoundError(Exception):
    pass


_token_user_cache: Dict[str, "UserAPI"] = {}


class UserAPI:
    def __init__(self, user: Union[str, UserAPIResponse], host: Optional[str] = None, auth: Optional[Any] = None):
        self._user_info: Optional[UserAPIResponse] = None
        if isinstance(user, UserAPIResponse):
            self._user_info = user
            self._username = user.username
        else:
            self._username = user
        self.host = host if host is not None else config.host

        if auth is None:
            self.auth = get_authenticator(host=host)
        else:
            self.auth = auth

    @staticmethod
    def get_user_from_token(token_or_authenticator: Union[str, Any], host: Optional[str] = None) -> "UserAPI":
        """
        Get a user associated with the token.

        This function has primitive caching for string tokens.
        """
        if host is None:
            host = config.host
        user_url = multi_urljoin(host, "api/v1/user")
        if isinstance(token_or_authenticator, str):
            if token_or_authenticator in _token_user_cache:
                return _token_user_cache[token_or_authenticator]
            auth = HTTPBearerAuth(token_or_authenticator)
        else:
            auth = token_or_authenticator
        resp = http_request("GET", user_url, auth=auth)
        if resp.status_code == 404:
            raise UserNotFoundError
        if resp.status_code != 200:
            raise RuntimeError(f"Got HTTP status {resp.status_code} while trying to get user: {resp.content}")
        user_info = dacite.from_dict(UserAPIResponse, resp.json())
        res = UserAPI(user=user_info, host=host, auth=auth)
        if isinstance(token_or_authenticator, str):
            _token_user_cache[token_or_authenticator] = res
        return res

    @staticmethod
    def get_current_user(host: Optional[str] = None) -> "UserAPI":
        return UserAPI.get_user_from_token(get_token(host=host), host=host)

    @property
    def username(self) -> str:
        return self.user_info.username

    @property
    def user_id(self) -> int:
        return self.user_info.id

    @cached_property
    def user_info(self) -> UserAPIResponse:
        if self._user_info is not None:
            return self._user_info
        user_url = multi_urljoin(self.host, "api/v1/users", self._username)
        resp = http_request("GET", user_url, auth=self.auth)
        if resp.status_code == 404:
            raise UserNotFoundError
        if resp.status_code != 200:
            raise RuntimeError(
                f"Got HTTP status {resp.status_code} while trying to get user {self._username}: {resp.content}"
            )
        user_info = dacite.from_dict(UserAPIResponse, resp.json())
        return user_info
